import logging
import uuid
from collections.abc import Iterable, Mapping
from typing import Protocol

import sqlalchemy as sa
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.permission.association_scopes_entities import (
    AssociationScopesEntitiesCreateInput,
)
from ai.backend.manager.data.permission.id import ObjectId, ScopeId
from ai.backend.manager.data.permission.object_permission import (
    ObjectPermissionCreateInput,
    ObjectPermissionData,
)
from ai.backend.manager.data.permission.permission import PermissionCreator, PermissionData
from ai.backend.manager.data.permission.permission_group import (
    PermissionGroupCreator,
    PermissionGroupData,
)
from ai.backend.manager.data.permission.types import (
    EntityType,
    OperationType,
    RoleSource,
)
from ai.backend.manager.data.permission.user_role import UserRoleCreateInput
from ai.backend.manager.models.rbac_models.association_scopes_entities import (
    AssociationScopesEntitiesRow,
)
from ai.backend.manager.models.rbac_models.permission.object_permission import ObjectPermissionRow
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow
from ai.backend.manager.models.rbac_models.permission.permission_group import PermissionGroupRow
from ai.backend.manager.models.rbac_models.user_role import UserRoleRow

from ...data.permission.role import (
    RoleCreateInput,
    RoleData,
)
from ...models.rbac_models.role import RoleRow

log = BraceStyleAdapter(logging.getLogger(__name__))


class ScopeSystemRoleData(Protocol):
    def scope_id(self) -> ScopeId: ...

    def role_name(self) -> str: ...

    def entity_operations(self) -> Mapping[EntityType, Iterable[OperationType]]:
        """Returns a mapping of entity types to the set of operations that should be granted for each entity type."""
        ...


class RoleManager:
    def __init__(self) -> None:
        pass

    async def create_system_role(
        self, db_session: SASession, data: ScopeSystemRoleData
    ) -> RoleData:
        role = await self._create_system_role(db_session, data)
        permission_group = await self._create_permission_group(db_session, data, role)
        await self._create_permissions(db_session, data, permission_group)
        return role

    async def _create_system_role(
        self, db_session: SASession, data: ScopeSystemRoleData
    ) -> RoleData:
        input = RoleCreateInput(
            name=data.role_name(),
            source=RoleSource.SYSTEM,
        )
        row = RoleRow.from_input(input)
        db_session.add(row)
        await db_session.flush()
        await db_session.refresh(row)
        return row.to_data()

    async def _create_permission_group(
        self, db_session: SASession, data: ScopeSystemRoleData, role: RoleData
    ) -> PermissionGroupData:
        input = PermissionGroupCreator(
            role_id=role.id,
            scope_id=data.scope_id(),
        )
        row = PermissionGroupRow.from_input(input)
        db_session.add(row)
        await db_session.flush()
        await db_session.refresh(row)
        return row.to_data()

    async def _create_permissions(
        self,
        db_session: SASession,
        data: ScopeSystemRoleData,
        permission_group: PermissionGroupData,
    ) -> list[PermissionData]:
        permission_rows: list[PermissionRow] = []
        for entity, operations in data.entity_operations().items():
            for operation in operations:
                creator = PermissionCreator(
                    permission_group_id=permission_group.id,
                    entity_type=entity,
                    operation=operation,
                )
                permission_rows.append(PermissionRow.from_input(creator))
        db_session.add_all(permission_rows)
        await db_session.flush()
        return [row.to_data() for row in permission_rows]

    async def map_user_to_role(
        self, db_session: SASession, user_id: uuid.UUID, role_id: uuid.UUID
    ) -> None:
        creator = UserRoleCreateInput(
            user_id=user_id,
            role_id=role_id,
        )
        await db_session.execute(sa.insert(UserRoleRow).values(creator.fields_to_store()))

    async def map_entity_to_scope(
        self,
        db_session: SASession,
        entity_id: ObjectId,
        scope_id: ScopeId,
    ) -> None:
        creator = AssociationScopesEntitiesCreateInput(
            scope_id=scope_id,
            object_id=entity_id,
        )
        try:
            await db_session.execute(
                sa.insert(AssociationScopesEntitiesRow).values(creator.fields_to_store())
            )
        except IntegrityError:
            log.exception(
                "entity and scope mapping already exists: {}, {}. Skipping.",
                entity_id.to_str(),
                scope_id.to_str(),
            )

    async def unmap_entity_from_scope(
        self,
        db_session: SASession,
        entity_id: ObjectId,
        scope_id: ScopeId,
    ) -> None:
        await db_session.execute(
            sa.delete(AssociationScopesEntitiesRow).where(
                sa.and_(
                    AssociationScopesEntitiesRow.scope_type == scope_id.scope_type,
                    AssociationScopesEntitiesRow.scope_id == scope_id.scope_id,
                    AssociationScopesEntitiesRow.entity_type == entity_id.entity_type,
                    AssociationScopesEntitiesRow.entity_id == entity_id.entity_id,
                )
            )
        )

    async def add_object_permission_to_user_role(
        self,
        db_session: SASession,
        user_id: uuid.UUID,
        entity_id: ObjectId,
        operations: Iterable[OperationType],
    ) -> list[ObjectPermissionData]:
        permission_group = await db_session.scalar(
            sa.select(PermissionGroupRow).where(PermissionGroupRow.scope_id == str(user_id))
        )
        role_id = permission_group.role_id

        creators = [
            ObjectPermissionCreateInput(
                role_id=role_id,
                entity_type=entity_id.entity_type,
                entity_id=entity_id.entity_id,
                operation=operation,
            )
            for operation in operations
        ]

        rows = [ObjectPermissionRow.from_input(creator) for creator in creators]
        db_session.add_all(rows)
        await db_session.flush()
        for row in rows:
            await db_session.refresh(row)
        result = [row.to_data() for row in rows]
        return result

    async def delete_object_permission_of_user(
        self,
        db_session: SASession,
        user_id: uuid.UUID,
        entity_id: uuid.UUID,
    ) -> None:
        permission_group = await db_session.scalar(
            sa.select(PermissionGroupRow).where(PermissionGroupRow.scope_id == str(user_id))
        )
        role_id = permission_group.role_id
        await db_session.execute(
            sa.delete(ObjectPermissionRow).where(
                sa.and_(
                    ObjectPermissionRow.role_id == role_id,
                    ObjectPermissionRow.entity_id == str(entity_id),
                )
            )
        )



import urllib
from collections.abc import Mapping
from contextvars import ContextVar
from typing import Any, Optional, Sequence, TypeAlias, override

import aiotools
import yarl

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.identity import get_instance_id
from ai.backend.common.types import SlotName, SlotTypes, current_resource_slots
from ai.backend.manager.api import ManagerStatus
from ai.backend.manager.config.loader.types import AbstractConfigLoader
from ai.backend.manager.defs import INTRINSIC_SLOTS
from ai.backend.manager.errors.common import ServerMisconfiguredError

current_vfolder_types: ContextVar[list[str]] = ContextVar("current_vfolder_types")


NestedStrKeyedDict: TypeAlias = "dict[str, Any | NestedStrKeyedDict]"


class LegacyEtcdLoader(AbstractConfigLoader):
    _etcd: AsyncEtcd
    _config_prefix: str = "config"

    def __init__(self, etcd: AsyncEtcd, config_prefix: Optional[str] = None) -> None:
        super().__init__()
        self._etcd = etcd
        if config_prefix:
            self._config_prefix = config_prefix

    @override
    async def load(self) -> Mapping[str, Any]:
        raw_cfg = await self._etcd.get_prefix(self._config_prefix)
        return raw_cfg

    async def close(self) -> None:
        await self._etcd.close()

    def __hash__(self) -> int:
        # When used as a key in dicts, we don't care our contents.
        # Just treat it like an opaque object.
        return hash(id(self))

    @classmethod
    def flatten(cls, key_prefix: str, inner_dict: NestedStrKeyedDict) -> dict[str, str]:
        flattened_dict: dict[str, str] = {}
        for k, v in inner_dict.items():
            if k == "":
                flattened_key = key_prefix
            else:
                flattened_key = key_prefix + "/" + urllib.parse.quote(k, safe="")
            match v:
                case Mapping():
                    flattened_dict.update(cls.flatten(flattened_key, v))  # type: ignore
                case str():
                    flattened_dict[flattened_key] = v
                case int() | float() | yarl.URL():
                    flattened_dict[flattened_key] = str(v)
                case _:
                    raise ValueError(
                        f"The value {v!r} must be serialized before storing to the etcd"
                    )
        return flattened_dict

    async def get_raw(self, key: str, allow_null: bool = True) -> Optional[str]:
        value = await self._etcd.get(key)
        if not allow_null and value is None:
            raise ServerMisconfiguredError("A required etcd config is missing.", key)
        return value

    async def register_myself(self) -> None:
        instance_id = await get_instance_id()
        manager_info = {
            f"nodes/manager/{instance_id}": "up",
        }
        await self._etcd.put_dict(manager_info)

    async def deregister_myself(self) -> None:
        instance_id = await get_instance_id()
        await self._etcd.delete_prefix(f"nodes/manager/{instance_id}")

    async def update_resource_slots(
        self,
        slot_key_and_units: Mapping[SlotName, SlotTypes],
    ) -> None:
        updates = {}
        known_slots = await self.get_resource_slots()
        for k, v in slot_key_and_units.items():
            if k not in known_slots or v != known_slots[k]:
                updates[f"config/resource_slots/{k}"] = v.value
        if updates:
            await self._etcd.put_dict(updates)

    async def update_manager_status(self, status) -> None:
        await self._etcd.put("manager/status", status.value)
        self.get_manager_status.cache_clear()

    @aiotools.lru_cache(maxsize=1, expire_after=2.0)
    async def _get_resource_slots(self):
        raw_data = await self._etcd.get_prefix_dict("config/resource_slots")
        return {SlotName(k): SlotTypes(v) for k, v in raw_data.items()}

    async def get_resource_slots(self) -> Mapping[SlotName, SlotTypes]:
        """
        Returns the system-wide known resource slots and their units.
        """
        try:
            ret = current_resource_slots.get()
        except LookupError:
            configured_slots = await self._get_resource_slots()
            ret = {**INTRINSIC_SLOTS, **configured_slots}
            current_resource_slots.set(ret)
        return ret

    @aiotools.lru_cache(maxsize=1, expire_after=2.0)
    async def _get_vfolder_types(self):
        return await self._etcd.get_prefix("volumes/_types")

    async def get_vfolder_types(self) -> Sequence[str]:
        """
        Returns the vfolder types currently set. One of "user" and/or "group".
        If none is specified, "user" type is implicitly assumed.
        """
        try:
            ret = current_vfolder_types.get()
        except LookupError:
            vf_types = await self._get_vfolder_types()
            ret = list(vf_types.keys())
            current_vfolder_types.set(ret)
        return ret

    @aiotools.lru_cache(maxsize=1, expire_after=5.0)
    async def get_manager_nodes_info(self):
        return await self._etcd.get_prefix_dict("nodes/manager")

    @aiotools.lru_cache(maxsize=1, expire_after=2.0)
    async def get_manager_status(self) -> ManagerStatus:
        status = await self._etcd.get("manager/status")
        if status is None:
            return ManagerStatus.TERMINATED
        return ManagerStatus(status)

    async def watch_manager_status(self):
        async with aiotools.aclosing(self._etcd.watch("manager/status")) as agen:
            async for ev in agen:
                yield ev

    # TODO: refactor using contextvars in Python 3.7 so that the result is cached
    #       in a per-request basis.
    @aiotools.lru_cache(maxsize=1, expire_after=2.0)
    async def get_allowed_origins(self):
        return await self._etcd.get("config/api/allow-origins")


class LegacyEtcdVolumesLoader(AbstractConfigLoader):
    _etcd: AsyncEtcd
    _config_prefix: str = "volumes"

    def __init__(self, etcd: AsyncEtcd) -> None:
        super().__init__()
        self._etcd = etcd

    @override
    async def load(self) -> Mapping[str, Any]:
        raw_cfg = await self._etcd.get_prefix(self._config_prefix)
        return {"volumes": raw_cfg}



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryCreatorMeta
from ai.backend.manager.data.reservoir_registry.creator import ReservoirRegistryCreator
from ai.backend.manager.data.reservoir_registry.types import ReservoirRegistryData
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class CreateReservoirRegistryAction(ArtifactRegistryAction):
    creator: ReservoirRegistryCreator
    meta: ArtifactRegistryCreatorMeta

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create_reservoir_registry"


@dataclass
class CreateReservoirActionResult(BaseActionResult):
    result: ReservoirRegistryData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.result.id)



import uuid

import sqlalchemy as sa
from sqlalchemy.orm import selectinload

from ai.backend.common.exception import (
    StorageNamespaceNotFoundError,
)
from ai.backend.common.metrics.metric import LayerType
from ai.backend.manager.data.object_storage.creator import ObjectStorageCreator
from ai.backend.manager.data.object_storage.modifier import ObjectStorageModifier
from ai.backend.manager.data.object_storage.types import ObjectStorageData
from ai.backend.manager.data.object_storage_namespace.types import StorageNamespaceData
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.errors.object_storage import (
    ObjectStorageNotFoundError,
)
from ai.backend.manager.models.object_storage import ObjectStorageRow
from ai.backend.manager.models.storage_namespace import StorageNamespaceRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

# Layer-specific decorator for user repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.OBJECT_STORAGE)


class ObjectStorageRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @repository_decorator()
    async def get_by_name(self, storage_name: str) -> ObjectStorageData:
        """
        Get an existing object storage configuration from the database.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(ObjectStorageRow).where(ObjectStorageRow.name == storage_name)
            result = await db_session.execute(query)
            row: ObjectStorageRow = result.scalar_one_or_none()
            if row is None:
                raise ObjectStorageNotFoundError(
                    f"Object storage with name {storage_name} not found."
                )
            return row.to_dataclass()

    @repository_decorator()
    async def get_by_id(self, storage_id: uuid.UUID) -> ObjectStorageData:
        """
        Get an existing object storage configuration from the database by ID.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(ObjectStorageRow).where(ObjectStorageRow.id == storage_id)
            result = await db_session.execute(query)
            row: ObjectStorageRow = result.scalar_one_or_none()
            if row is None:
                raise ObjectStorageNotFoundError(f"Object storage with ID {storage_id} not found.")
            return row.to_dataclass()

    @repository_decorator()
    async def get_by_namespace_id(self, storage_namespace_id: uuid.UUID) -> ObjectStorageData:
        """
        Get an existing object storage configuration from the database by ID.
        """
        async with self._db.begin_session() as db_session:
            query = (
                sa.select(StorageNamespaceRow)
                .where(StorageNamespaceRow.id == storage_namespace_id)
                .options(selectinload(StorageNamespaceRow.object_storage_row))
            )
            result = await db_session.execute(query)
            row: StorageNamespaceRow = result.scalar_one_or_none()
            if row is None:
                raise ObjectStorageNotFoundError(
                    f"Object storage with namespace ID {storage_namespace_id} not found."
                )
            return row.object_storage_row.to_dataclass()

    @repository_decorator()
    async def get_storage_namespace(
        self, storage_id: uuid.UUID, bucket_name: str
    ) -> StorageNamespaceData:
        """
        Get an existing object storage namespace from the database.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(StorageNamespaceRow).where(
                StorageNamespaceRow.storage_id == storage_id,
                StorageNamespaceRow.namespace == bucket_name,
            )
            result = await db_session.execute(query)
            row: StorageNamespaceRow = result.scalar_one_or_none()
            if row is None:
                raise StorageNamespaceNotFoundError(
                    f"Object storage namespace with bucket_name {bucket_name} not found."
                )
            return row.to_dataclass()

    @repository_decorator()
    async def get_storage_namespace_by_id(
        self, storage_namespace_id: uuid.UUID
    ) -> StorageNamespaceData:
        """
        Get an existing object storage namespace from the database by ID.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(StorageNamespaceRow).where(
                StorageNamespaceRow.id == storage_namespace_id
            )
            result = await db_session.execute(query)
            row: StorageNamespaceRow = result.scalar_one_or_none()
            if row is None:
                raise StorageNamespaceNotFoundError(
                    f"Object storage namespace ID {storage_namespace_id} not found."
                )
            return row.to_dataclass()

    @repository_decorator()
    async def create(self, creator: ObjectStorageCreator) -> ObjectStorageData:
        """
        Create a new object storage configuration in the database.
        """
        async with self._db.begin_session() as db_session:
            object_storage_data = creator.fields_to_store()
            object_storage_row = ObjectStorageRow(**object_storage_data)
            db_session.add(object_storage_row)
            await db_session.flush()
            await db_session.refresh(object_storage_row)
            return object_storage_row.to_dataclass()

    @repository_decorator()
    async def update(
        self, storage_id: uuid.UUID, modifier: ObjectStorageModifier
    ) -> ObjectStorageData:
        """
        Update an existing object storage configuration in the database.
        """
        async with self._db.begin_session() as db_session:
            data = modifier.fields_to_update()
            update_stmt = (
                sa.update(ObjectStorageRow)
                .where(ObjectStorageRow.id == storage_id)
                .values(**data)
                .returning(*sa.select(ObjectStorageRow).selected_columns)
            )
            stmt = sa.select(ObjectStorageRow).from_statement(update_stmt)
            row: ObjectStorageRow = (await db_session.execute(stmt)).scalars().one()

            return row.to_dataclass()

    @repository_decorator()
    async def delete(self, storage_id: uuid.UUID) -> uuid.UUID:
        """
        Delete an existing object storage configuration from the database.
        """
        async with self._db.begin_session() as db_session:
            delete_query = (
                sa.delete(ObjectStorageRow)
                .where(ObjectStorageRow.id == storage_id)
                .returning(ObjectStorageRow.id)
            )
            result = await db_session.execute(delete_query)
            deleted_id = result.scalar()
            return deleted_id

    @repository_decorator()
    async def list_object_storages(self) -> list[ObjectStorageData]:
        """
        List all object storage configurations from the database.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(ObjectStorageRow)
            result = await db_session.execute(query)
            rows: list[ObjectStorageRow] = result.scalars().all()
            return [row.to_dataclass() for row in rows]



"""
Data types returned by ScheduleCacheSource.
These types include transformation methods to convert to common entities.
"""

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ai.backend.common.types import AccessKey
from ai.backend.manager.sokovan.scheduler.types import ConcurrencySnapshot

if TYPE_CHECKING:
    pass


@dataclass
class CacheConcurrencyData:
    """Concurrency data fetched from cache (Redis/Valkey)."""

    sessions_by_keypair: dict[AccessKey, int]
    sftp_sessions_by_keypair: dict[AccessKey, int]

    def to_concurrency_snapshot(self) -> ConcurrencySnapshot:
        """Convert to ConcurrencySnapshot entity."""
        return ConcurrencySnapshot(
            sessions_by_keypair=self.sessions_by_keypair,
            sftp_sessions_by_keypair=self.sftp_sessions_by_keypair,
        )


@dataclass
class CacheConcurrencyUpdate:
    """Concurrency updates to be applied to cache."""

    regular: dict[AccessKey, int]  # Regular session concurrency changes
    sftp: dict[AccessKey, int]  # SFTP session concurrency changes



from typing import Any
from uuid import UUID

import aiohttp

from ai.backend.common.metrics.metric import LayerType
from ai.backend.manager.decorators.client_decorator import create_layer_aware_client_decorator

from .types import CreateEndpointRequestBody

client_decorator = create_layer_aware_client_decorator(LayerType.WSPROXY_CLIENT)


class AppProxyClient:
    _client_session: aiohttp.ClientSession
    _address: str
    _token: str

    def __init__(self, client_session: aiohttp.ClientSession, address: str, token: str) -> None:
        self._client_session = client_session
        self._address = address
        self._token = token

    @client_decorator()
    async def create_endpoint(
        self,
        endpoint_id: UUID,
        body: CreateEndpointRequestBody,
    ) -> dict[str, Any]:
        async with self._client_session.post(
            f"/v2/endpoints/{endpoint_id}",
            json=body.model_dump(mode="json"),
            headers={
                "X-BackendAI-Token": self._token,
            },
        ) as resp:
            resp.raise_for_status()
            return await resp.json()

    @client_decorator()
    async def delete_endpoint(
        self,
        endpoint_id: UUID,
    ) -> None:
        async with self._client_session.delete(
            f"/v2/endpoints/{endpoint_id}",
            headers={
                "X-BackendAI-Token": self._token,
            },
        ):
            pass



from dataclasses import dataclass
from uuid import UUID

from ai.backend.manager.data.deployment.types import (
    DeploymentMetadata,
    DeploymentNetworkSpec,
    ModelRevisionSpec,
    ReplicaSpec,
)
from ai.backend.manager.data.image.types import ImageIdentifier


@dataclass
class DeploymentCreator:
    metadata: DeploymentMetadata
    replica_spec: ReplicaSpec
    network: DeploymentNetworkSpec
    model_revision: ModelRevisionSpec

    # Accessor properties for backward compatibility
    @property
    def image_identifier(self) -> ImageIdentifier:
        """Get the image identifier from model revision spec."""
        return self.model_revision.image_identifier

    @property
    def domain(self) -> str:
        """Get the domain name from metadata."""
        return self.metadata.domain

    @property
    def project(self) -> UUID:
        """Get the project ID from metadata."""
        return self.metadata.project

    @property
    def name(self) -> str:
        """Get the deployment name from metadata."""
        return self.metadata.name



from typing import Self

import sqlalchemy as sa

from ai.backend.common.events.types import AbstractEvent, EventDomain

from .base import (
    Base,
    IDColumn,
    StrEnumType,
)

__all__ = ("EventLogRow",)


class EventLogRow(Base):
    __tablename__ = "event_logs"

    id = IDColumn("id")

    event_name = sa.Column("event_name", sa.String, nullable=False)
    event_domain = sa.Column(
        "event_domain",
        StrEnumType(EventDomain),
        nullable=False,
    )
    domain_id = sa.Column("domain_id", sa.String, nullable=True, index=True)

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        nullable=False,
        index=True,
    )
    duration = sa.Column("duration", sa.Interval, nullable=True)

    @classmethod
    def from_event(cls, event: AbstractEvent) -> Self:
        event_name = event.event_name()
        event_domain = event.event_domain()
        event_domain_id = event.domain_id()
        return EventLogRow(
            event_name=event_name,
            event_domain=event_domain,
            domain_id=event_domain_id,
        )



import enum

from ai.backend.manager.data.permission.status import (
    RoleStatus as OriginalRoleStatus,
)
from ai.backend.manager.data.permission.types import (
    EntityType as OriginalEntityType,
)
from ai.backend.manager.data.permission.types import (
    OperationType as OriginalOperationType,
)
from ai.backend.manager.data.permission.types import (
    RoleSource as OriginalRoleSource,
)
from ai.backend.manager.data.permission.types import (
    ScopeType as OriginalScopeType,
)


class RoleSource(enum.StrEnum):
    """
    Definition source of the role.
    """

    SYSTEM = "system"  # System-defined role, e.g., default roles
    CUSTOM = "custom"  # Custom role defined

    def to_original(self) -> OriginalRoleSource:
        return OriginalRoleSource(self.value)


class RoleStatus(enum.StrEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    DELETED = "deleted"

    def to_original(self) -> OriginalRoleStatus:
        return OriginalRoleStatus(self.value)


class OperationType(enum.StrEnum):
    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    SOFT_DELETE = "soft-delete"
    HARD_DELETE = "hard-delete"
    GRANT_ALL = "grant:all"  # Allow user to grant all permissions, including grant of grant
    GRANT_READ = "grant:read"
    GRANT_UPDATE = "grant:update"
    GRANT_SOFT_DELETE = "grant:soft-delete"
    GRANT_HARD_DELETE = "grant:hard-delete"

    def to_original(self) -> OriginalOperationType:
        return OriginalOperationType(self.value)

    @classmethod
    def owner_operations(cls) -> set["OperationType"]:
        """
        Returns a set of operations that are considered owner operations.
        Owner operations are those that allow full control over an entity.
        """
        return {op for op in cls}

    @classmethod
    def admin_operations(cls) -> set["OperationType"]:
        """
        Returns a set of operations that are considered admin operations.
        Admin operations are those that allow management of entities, including creation and deletion.
        """
        return {op for op in cls}

    @classmethod
    def member_operations(cls) -> set["OperationType"]:
        """
        Returns a set of operations that are considered member operations.
        Member operations are those that allow read access.
        """
        return {
            cls.READ,
        }


class ScopeType(enum.StrEnum):
    DOMAIN = "domain"
    PROJECT = "project"
    USER = "user"

    def to_original(self) -> OriginalScopeType:
        return OriginalScopeType(self.value)


class EntityType(enum.StrEnum):
    USER = "user"
    PROJECT = "project"
    DOMAIN = "domain"

    VFOLDER = "vfolder"
    IMAGE = "image"
    SESSION = "session"

    def to_original(self) -> OriginalEntityType:
        return OriginalEntityType(self.value)

    @classmethod
    def _scope_types(cls) -> set["EntityType"]:
        """
        Returns a set of entity types that are considered scope types.
        """
        return {cls.USER, cls.PROJECT, cls.DOMAIN}

    @classmethod
    def _resource_types(cls) -> set["EntityType"]:
        """
        Returns a set of entity types that are considered resource types.
        """
        return {
            cls.VFOLDER,
            cls.IMAGE,
            cls.SESSION,
        }

    @classmethod
    def owner_accessible_entity_types_in_user(cls) -> set["EntityType"]:
        """
        Returns a set of entity types that are accessible by owner roles in user scope.
        """
        return cls._resource_types()

    @classmethod
    def admin_accessible_entity_types_in_project(cls) -> set["EntityType"]:
        """
        Returns a set of entity types that are accessible by admin roles.
        """
        return {*cls._resource_types(), cls.USER}

    @classmethod
    def member_accessible_entity_types_in_project(cls) -> set["EntityType"]:
        """
        Returns a set of entity types that are accessible by member roles.
        """
        return {*cls._resource_types(), cls.USER}



from uuid import UUID

from sqlalchemy.exc import DBAPIError
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.docker import ImageRef
from ai.backend.common.types import ImageAlias
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.errors.image import (
    ImageNotFound,
    PurgeImageActionByIdObjectDBError,
)
from ai.backend.manager.models.image import (
    ImageIdentifier,
    ImageRow,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine


class AdminImageRepository:
    """
    Repository for admin-specific image operations that bypass ownership checks.
    This should only be used by superadmin users.
    """

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def _resolve_image(
        self,
        session: SASession,
        identifiers: list[ImageAlias | ImageRef | ImageIdentifier],
    ) -> ImageRow:
        return await ImageRow.resolve(session, identifiers)

    async def _get_image_by_id(
        self, session: SASession, image_id: UUID, load_aliases: bool = False
    ) -> ImageRow:
        image_row = await ImageRow.get(session, image_id, load_aliases=load_aliases)
        if not image_row:
            raise ImageNotFound()
        return image_row

    async def soft_delete_image_force(
        self, identifiers: list[ImageAlias | ImageRef | ImageIdentifier]
    ) -> ImageData:
        """
        Marks an image as deleted without checking ownership.
        This is a forceful deletion and should be used with caution.
        """
        async with self._db.begin_session() as session:
            row = await self._resolve_image(session, identifiers)
            await row.mark_as_deleted(session)
            data = row.to_dataclass()
        return data

    async def soft_delete_image_by_id_force(self, image_id: UUID) -> ImageData:
        """
        Marks an image as deleted by its ID without checking ownership.
        This is a forceful deletion and should be used with caution.
        """
        async with self._db.begin_session() as session:
            image_row = await self._get_image_by_id(session, image_id)
            await image_row.mark_as_deleted(session)
            data = image_row.to_dataclass()
        return data

    async def delete_image_with_aliases_force(self, image_id: UUID) -> ImageData:
        """
        Deletes an image and all its aliases without checking ownership.
        This is a forceful deletion and should be used with caution.
        """
        try:
            async with self._db.begin_session() as session:
                image_row = await self._get_image_by_id(session, image_id, load_aliases=True)
                data = image_row.to_dataclass()
                for alias in image_row.aliases:
                    await session.delete(alias)
                await session.delete(image_row)
            return data
        except DBAPIError as e:
            raise PurgeImageActionByIdObjectDBError(str(e))

    async def untag_image_from_registry_force(self, image_id: UUID) -> ImageData:
        """
        Untags an image from registry without ownership check.
        This is an admin-only operation.
        """
        async with self._db.begin_readonly_session() as session:
            image_row = await self._get_image_by_id(session, image_id, load_aliases=True)
            await image_row.untag_image_from_registry(self._db, session)
            data = image_row.to_dataclass()
        return data



import uuid
from dataclasses import dataclass
from typing import Any

from ai.backend.manager.data.permission.id import ObjectId, ScopeId

from .enums import RoleSource, RoleStatus, ScopeType

ROLE_NAME_PREFIX = "role_"
ADMIN_ROLE_NAME_SUFFIX = "_admin"


@dataclass
class UserRoleCreateInput:
    user_id: uuid.UUID
    role_id: uuid.UUID

    def to_dict(self) -> dict[str, uuid.UUID]:
        return {
            "user_id": self.user_id,
            "role_id": self.role_id,
        }


@dataclass
class AssociationScopesEntitiesCreateInput:
    scope_id: ScopeId
    object_id: ObjectId

    def to_dict(self) -> dict[str, Any]:
        return {
            "scope_type": self.scope_id.scope_type,
            "scope_id": self.scope_id.scope_id,
            "entity_type": self.object_id.entity_type,
            "entity_id": self.object_id.entity_id,
        }


@dataclass
class RoleCreateInput:
    name: str
    source: RoleSource
    status: RoleStatus = RoleStatus.ACTIVE

    def to_dict(self) -> dict[str, str]:
        return {
            "name": self.name,
            "source": self.source,
            "status": self.status,
        }


@dataclass
class PermissionGroupCreateInput:
    role_id: uuid.UUID
    scope_type: ScopeType
    scope_id: str

    def to_dict(self) -> dict[str, Any]:
        return {
            "role_id": self.role_id,
            "scope_type": self.scope_type,
            "scope_id": self.scope_id,
        }


@dataclass
class PermissionCreateInput:
    permission_group_id: uuid.UUID
    entity_type: str
    operation: str

    def to_dict(self) -> dict[str, Any]:
        return {
            "permission_group_id": self.permission_group_id,
            "entity_type": self.entity_type,
            "operation": self.operation,
        }


@dataclass
class ObjectPermissionCreateInput:
    role_id: uuid.UUID
    entity_id: str
    entity_type: str
    operation: str

    def to_dict(self) -> dict[str, Any]:
        return {
            "role_id": self.role_id,
            "entity_id": self.entity_id,
            "entity_type": self.entity_type,
            "operation": self.operation,
        }



from __future__ import annotations

import asyncio
import logging
import os
import socket
from typing import TYPE_CHECKING, Optional, cast

import redis.exceptions
from pydantic import (
    BaseModel,
    Field,
)
from sqlalchemy.pool import Pool

from ai.backend.common import msgpack
from ai.backend.logging import BraceStyleAdapter

if TYPE_CHECKING:
    from ..api.context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__: tuple[str, ...] = (
    "SQLAlchemyConnectionInfo",
    "RedisObjectConnectionInfo",
    "get_sqlalchemy_connection_info",
    "get_redis_object_info_list",
    "_get_connnection_info",
    "report_manager_status",
)

_sqlalchemy_pool_type_names = (
    "AssertionPool",
    "AsyncAdaptedQueuePool",
    "FallbackAsyncAdaptedQueuePool",
    "NullPool",
    "QueuePool",
    "SingletonThreadPool",
    "StaticPool",
)

MANAGER_STATUS_KEY = "manager.status"


def _get_connection_status_key(node_id: str, pid: int) -> str:
    return f"{MANAGER_STATUS_KEY}.{node_id}:{pid}"


class SQLAlchemyConnectionInfo(BaseModel):
    pool_type: str = Field(
        description=f"Connection pool type of SQLAlchemy engine. One of {_sqlalchemy_pool_type_names}.",
    )
    status_description: str
    num_checkedout_cxn: int = Field(
        description="The number of open connections in SQLAlchemy connection pool.",
    )
    num_checkedin_cxn: int = Field(
        description="The number of closed connections in SQLAlchemy connection pool.",
    )

    @property
    def total_cxn(self) -> int:
        return self.num_checkedout_cxn + self.num_checkedin_cxn


class RedisObjectConnectionInfo(BaseModel):
    name: str
    num_connections: Optional[int] = Field(
        description="The number of connections in Redis Client's connection pool."
    )
    max_connections: int
    err_msg: Optional[str] = Field(
        description="Error message occurred when fetch connection info from Redis client objects.",
        default=None,
    )


class ConnectionInfoOfProcess(BaseModel):
    node_id: str = Field(description="Specified Manager ID or hostname.")
    pid: int = Field(description="Process ID.")
    sqlalchemy_info: SQLAlchemyConnectionInfo
    redis_connection_info: list[RedisObjectConnectionInfo]


async def get_sqlalchemy_connection_info(root_ctx: RootContext) -> SQLAlchemyConnectionInfo:
    pool = cast(Pool, root_ctx.db.pool)
    sqlalchemy_info = SQLAlchemyConnectionInfo(
        pool_type=type(pool).__name__,
        status_description=pool.status(),
        num_checkedout_cxn=pool.checkedout(),
        num_checkedin_cxn=pool.checkedin(),
    )
    return sqlalchemy_info


async def get_redis_object_info_list(root_ctx: RootContext) -> list[RedisObjectConnectionInfo]:
    log.warning("get_redis_object_info_list is deprecated.")
    return []


async def _get_connnection_info(root_ctx: RootContext) -> ConnectionInfoOfProcess:
    node_id = root_ctx.config_provider.config.manager.id or socket.gethostname()
    pid = os.getpid()

    sqlalchemy_info = await get_sqlalchemy_connection_info(root_ctx)
    redis_infos = await get_redis_object_info_list(root_ctx)
    return ConnectionInfoOfProcess(
        node_id=node_id, pid=pid, sqlalchemy_info=sqlalchemy_info, redis_connection_info=redis_infos
    )


async def report_manager_status(root_ctx: RootContext) -> None:
    lifetime = root_ctx.config_provider.config.manager.status_lifetime
    cxn_info = await _get_connnection_info(root_ctx)
    _data = msgpack.packb(cxn_info.model_dump(mode="json"))

    if lifetime is not None:
        await root_ctx.valkey_stat.set_manager_status(
            node_id=cxn_info.node_id,
            pid=cxn_info.pid,
            status_data=_data,
            lifetime=lifetime,
        )


async def get_manager_db_cxn_status(root_ctx: RootContext) -> list[ConnectionInfoOfProcess]:
    cxn_infos: list[ConnectionInfoOfProcess] = []

    try:
        _raw_value = cast(
            list[bytes] | None,
            await root_ctx.valkey_stat.scan_and_get_manager_status(
                f"{MANAGER_STATUS_KEY}*",
            ),
        )
    except (asyncio.TimeoutError, redis.exceptions.ConnectionError):
        # Cannot get data from redis. Return process's own info.
        cxn_infos = [(await _get_connnection_info(root_ctx))]
    else:
        if _raw_value is not None:
            cxn_infos = [
                ConnectionInfoOfProcess.model_validate(msgpack.unpackb(val)) for val in _raw_value
            ]
        else:
            cxn_infos = [(await _get_connnection_info(root_ctx))]
    return cxn_infos



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.domain.actions.create_domain import (
    CreateDomainAction,
    CreateDomainActionResult,
)
from ai.backend.manager.services.domain.actions.create_domain_node import (
    CreateDomainNodeAction,
    CreateDomainNodeActionResult,
)
from ai.backend.manager.services.domain.actions.delete_domain import (
    DeleteDomainAction,
    DeleteDomainActionResult,
)
from ai.backend.manager.services.domain.actions.modify_domain import (
    ModifyDomainAction,
    ModifyDomainActionResult,
)
from ai.backend.manager.services.domain.actions.modify_domain_node import (
    ModifyDomainNodeAction,
    ModifyDomainNodeActionResult,
)
from ai.backend.manager.services.domain.actions.purge_domain import (
    PurgeDomainAction,
    PurgeDomainActionResult,
)

from .service import DomainService


class DomainProcessors(AbstractProcessorPackage):
    create_domain_node: ActionProcessor[CreateDomainNodeAction, CreateDomainNodeActionResult]
    modify_domain_node: ActionProcessor[ModifyDomainNodeAction, ModifyDomainNodeActionResult]
    create_domain: ActionProcessor[CreateDomainAction, CreateDomainActionResult]
    modify_domain: ActionProcessor[ModifyDomainAction, ModifyDomainActionResult]
    delete_domain: ActionProcessor[DeleteDomainAction, DeleteDomainActionResult]
    purge_domain: ActionProcessor[PurgeDomainAction, PurgeDomainActionResult]

    def __init__(self, service: DomainService, action_monitors: list[ActionMonitor]) -> None:
        self.create_domain_node = ActionProcessor(service.create_domain_node, action_monitors)
        self.modify_domain_node = ActionProcessor(service.modify_domain_node, action_monitors)
        self.create_domain = ActionProcessor(service.create_domain, action_monitors)
        self.modify_domain = ActionProcessor(service.modify_domain, action_monitors)
        self.delete_domain = ActionProcessor(service.delete_domain, action_monitors)
        self.purge_domain = ActionProcessor(service.purge_domain, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateDomainNodeAction.spec(),
            ModifyDomainNodeAction.spec(),
            CreateDomainAction.spec(),
            ModifyDomainAction.spec(),
            DeleteDomainAction.spec(),
            PurgeDomainAction.spec(),
        ]



import uuid
from typing import Any, Mapping, Optional, Sequence, Union

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import selectinload

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.metrics.metric import LayerType
from ai.backend.common.types import VFolderHostPermission, VFolderID
from ai.backend.manager.data.permission.id import ObjectId, ScopeId
from ai.backend.manager.data.permission.types import EntityType, OperationType, ScopeType
from ai.backend.manager.data.vfolder.types import (
    VFolderAccessInfo,
    VFolderCreateParams,
    VFolderData,
    VFolderInvitationData,
    VFolderListResult,
    VFolderPermissionData,
)
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.errors.common import ObjectNotFound
from ai.backend.manager.errors.resource import ProjectNotFound
from ai.backend.manager.errors.storage import (
    VFolderDeletionNotAllowed,
    VFolderNotFound,
)
from ai.backend.manager.errors.user import UserNotFound
from ai.backend.manager.models.group import GroupRow, ProjectType
from ai.backend.manager.models.keypair import KeyPairRow
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import UserRole, UserRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, execute_with_retry
from ai.backend.manager.models.vfolder import (
    VFolderCloneInfo,
    VFolderDeletionInfo,
    VFolderInvitationRow,
    VFolderInvitationState,
    VFolderOperationStatus,
    VFolderOwnershipType,
    VFolderPermission,
    VFolderPermissionRow,
    VFolderRow,
    delete_vfolder_relation_rows,
    ensure_host_permission_allowed,
    get_sessions_by_mounted_folder,
    is_unmanaged,
    query_accessible_vfolders,
    vfolders,
)
from ai.backend.manager.services.vfolder.exceptions import VFolderInvalidParameter

from ..permission_controller.role_manager import RoleManager

# Layer-specific decorator for vfolder repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.VFOLDER)


class VfolderRepository:
    _db: ExtendedAsyncSAEngine
    _role_manager: RoleManager

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db
        self._role_manager = RoleManager()

    @repository_decorator()
    async def get_by_id_validated(
        self, vfolder_id: uuid.UUID, user_id: uuid.UUID, domain_name: str
    ) -> VFolderData:
        """
        Get a VFolder by ID with ownership/permission validation.
        Returns VFolderData if user has access.
        Raises VFolderNotFound if vfolder doesn't exist or user has no access.
        """
        async with self._db.begin_session() as session:
            vfolder_row = await self._get_vfolder_by_id(session, vfolder_id)
            if not vfolder_row:
                raise VFolderNotFound()

            # Check access permissions
            user_row = await session.scalar(sa.select(UserRow).where(UserRow.uuid == user_id))
            if not user_row:
                raise ObjectNotFound(object_name="User")

            # Check if user has access to this vfolder
            allowed_vfolder_types = ["user", "group"]  # TODO: get from config
            vfolder_dicts = await query_accessible_vfolders(
                session.bind,
                user_id,
                allow_privileged_access=True,
                user_role=user_row.role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=(VFolderRow.id == vfolder_id),
            )

            if not vfolder_dicts:
                raise VFolderNotFound()

            return self._vfolder_row_to_data(vfolder_row)

    @repository_decorator()
    async def get_by_id(self, vfolder_id: uuid.UUID) -> Optional[VFolderData]:
        """
        Get a VFolder by ID without validation.
        Returns VFolderData if found, None otherwise.
        """
        async with self._db.begin_session() as session:
            vfolder_row = await self._get_vfolder_by_id(session, vfolder_id)
            if not vfolder_row:
                return None
            return self._vfolder_row_to_data(vfolder_row)

    @repository_decorator()
    async def get_allowed_vfolder_hosts(
        self, user_uuid: uuid.UUID, group_uuid: Optional[uuid.UUID]
    ) -> str:
        """
        Get the allowed VFolder hosts for a user.
        """
        async with self._db.begin_readonly_session() as db_session:
            if group_uuid:
                group_row: Optional[GroupRow] = await db_session.scalar(
                    sa.select(GroupRow).where(GroupRow.id == group_uuid)
                )
                if group_row is None:
                    raise ProjectNotFound(f"Project with {group_uuid} not found.")

                return group_row.allowed_vfolder_hosts

            user_row: Optional[UserRow] = await db_session.scalar(
                sa.select(UserRow)
                .where(UserRow.uuid == user_uuid)
                .options(
                    selectinload(UserRow.main_keypair).selectinload(KeyPairRow.resource_policy_row)
                )
            )
            if user_row is None:
                raise UserNotFound(f"User with UUID {user_uuid} not found.")

            return user_row.main_keypair.resource_policy_row.allowed_vfolder_hosts

    @repository_decorator()
    async def get_max_vfolder_count(
        self, user_uuid: uuid.UUID, group_uuid: Optional[uuid.UUID]
    ) -> int:
        """
        Get the maximum VFolder count for a user or group.
        """
        async with self._db.begin_readonly_session() as db_session:
            if group_uuid:
                group_row: Optional[GroupRow] = await db_session.scalar(
                    sa.select(GroupRow)
                    .where(GroupRow.id == group_uuid)
                    .options(selectinload(GroupRow.resource_policy_row))
                )
                if group_row is None:
                    raise ProjectNotFound(f"Project with {group_uuid} not found.")

                return group_row.resource_policy_row.max_vfolder_count

            user_row: Optional[UserRow] = await db_session.scalar(
                sa.select(UserRow)
                .where(UserRow.uuid == user_uuid)
                .options(selectinload(UserRow.resource_policy_row))
            )
            if user_row is None:
                raise UserNotFound(f"User with UUID {user_uuid} not found.")

            return user_row.resource_policy_row.max_vfolder_count

    @repository_decorator()
    async def list_accessible_vfolders(
        self,
        user_id: uuid.UUID,
        user_role: UserRole,
        domain_name: str,
        allowed_vfolder_types: list[str],
        extra_conditions: Optional[sa.sql.expression.BinaryExpression] = None,
    ) -> VFolderListResult:
        """
        List all VFolders accessible to a user.
        Returns VFolderListResult with access information.
        """
        async with self._db.begin_session() as session:
            vfolder_dicts = await query_accessible_vfolders(
                session.bind,
                user_id,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=extra_conditions,
            )

            vfolder_access_infos = []
            for vfolder_dict in vfolder_dicts:
                vfolder_data = self._vfolder_dict_to_data(dict(vfolder_dict))
                is_owner = vfolder_dict.get("is_owner", False)
                permission = vfolder_dict.get("permission", VFolderPermission.READ_ONLY)

                vfolder_access_infos.append(
                    VFolderAccessInfo(
                        vfolder_data=vfolder_data,
                        is_owner=is_owner,
                        effective_permission=permission,
                    )
                )

            return VFolderListResult(vfolders=vfolder_access_infos)

    @repository_decorator()
    async def create_vfolder_with_permission(
        self, params: VFolderCreateParams, create_owner_permission: bool = False
    ) -> VFolderData:
        """
        Create a new VFolder with the given parameters and optionally create owner permission.
        Returns the created VFolderData.
        """
        async with self._db.begin_session() as session:
            # Create the VFolder
            insert_values = {
                "id": params.id.hex,
                "name": params.name,
                "domain_name": params.domain_name,
                "quota_scope_id": params.quota_scope_id,
                "usage_mode": params.usage_mode,
                "permission": params.permission,
                "last_used": None,
                "host": params.host,
                "creator": params.creator,
                "ownership_type": params.ownership_type,
                "user": params.user,
                "group": params.group,
                "unmanaged_path": params.unmanaged_path,
                "cloneable": params.cloneable,
                "status": params.status,
            }

            query = sa.insert(VFolderRow, insert_values)
            result = await session.execute(query)
            assert result.rowcount == 1
            match params.ownership_type:
                case VFolderOwnershipType.USER:
                    scope_id = ScopeId(ScopeType.USER, str(params.user))
                case VFolderOwnershipType.GROUP:
                    scope_id = ScopeId(ScopeType.PROJECT, str(params.group))
            await self._role_manager.map_entity_to_scope(
                session,
                entity_id=ObjectId(
                    entity_type=EntityType.VFOLDER,
                    entity_id=str(params.id),
                ),
                scope_id=scope_id,
            )

            # Create owner permission if requested
            if create_owner_permission and params.user:
                permission_insert = sa.insert(VFolderPermissionRow).values({
                    "user": params.user,
                    "vfolder": params.id.hex,
                    "permission": VFolderPermission.OWNER_PERM,
                })
                await session.execute(permission_insert)
                await self._role_manager.map_entity_to_scope(
                    session,
                    entity_id=ObjectId(
                        entity_type=EntityType.VFOLDER,
                        entity_id=str(params.id),
                    ),
                    scope_id=ScopeId(ScopeType.USER, str(params.user)),
                )
                await self._role_manager.add_object_permission_to_user_role(
                    session,
                    user_id=params.user,
                    entity_id=ObjectId(
                        entity_type=EntityType.VFOLDER,
                        entity_id=str(params.id),
                    ),
                    operations=[OperationType.READ],
                )

            # Return the created vfolder data
            created_vfolder = await self._get_vfolder_by_id(session, params.id)
            if not created_vfolder:
                raise VFolderNotFound()
            return self._vfolder_row_to_data(created_vfolder)

    @repository_decorator()
    async def update_vfolder_attribute(
        self, vfolder_id: uuid.UUID, field_updates: dict[str, Any]
    ) -> VFolderData:
        """
        Update VFolder attributes.
        Returns updated VFolderData.
        """
        async with self._db.begin_session() as session:
            vfolder_row = await self._get_vfolder_by_id(session, vfolder_id)
            if not vfolder_row:
                raise VFolderNotFound()

            for key, value in field_updates.items():
                if hasattr(vfolder_row, key):
                    setattr(vfolder_row, key, value)

            await session.flush()
            return self._vfolder_row_to_data(vfolder_row)

    @repository_decorator()
    async def move_vfolders_to_trash(self, vfolder_ids: list[uuid.UUID]) -> list[VFolderData]:
        """
        Move VFolders to trash.
        """

        async with self._db.begin_session() as session:
            vfolder_rows = []
            for vfolder_id in vfolder_ids:
                vfolder_row = await self._get_vfolder_by_id(session, vfolder_id)
                if vfolder_row:
                    vfolder_rows.append(vfolder_row)

            # Create deletion info for each vfolder
            deletion_infos = []
            for vfolder_row in vfolder_rows:
                vfolder_id_obj = VFolderID(
                    quota_scope_id=vfolder_row.quota_scope_id,
                    folder_id=vfolder_row.id,
                )
                deletion_info = VFolderDeletionInfo(
                    vfolder_id=vfolder_id_obj,
                    host=vfolder_row.host,
                    unmanaged_path=vfolder_row.unmanaged_path,
                )
                deletion_infos.append(deletion_info)

            # Note: initiate_vfolder_deletion requires storage_manager parameter
            # This would need to be passed to the repository method or handled differently
            # For now, we'll update the status directly instead of using the full deletion process
            for vfolder_row in vfolder_rows:
                mount_sessions = await get_sessions_by_mounted_folder(
                    session, VFolderID.from_row(vfolder_row)
                )
                if mount_sessions:
                    session_ids = [str(session_id) for session_id in mount_sessions]
                    raise VFolderDeletionNotAllowed(
                        "Cannot delete the vfolder. "
                        f"The vfolder(id: {vfolder_row.id}) is mounted on sessions(ids: {session_ids})."
                    )
                vfolder_row.status = VFolderOperationStatus.DELETE_PENDING

            await session.flush()

            return [self._vfolder_row_to_data(row) for row in vfolder_rows]

    @repository_decorator()
    async def restore_vfolders_from_trash(self, vfolder_ids: list[uuid.UUID]) -> list[VFolderData]:
        """
        Restore VFolders from trash.
        """
        async with self._db.begin_session() as session:
            vfolder_rows = []
            for vfolder_id in vfolder_ids:
                vfolder_row = await self._get_vfolder_by_id(session, vfolder_id)
                if vfolder_row:
                    vfolder_row.status = VFolderOperationStatus.READY
                    vfolder_rows.append(vfolder_row)

            await session.flush()
            return [self._vfolder_row_to_data(row) for row in vfolder_rows]

    @repository_decorator()
    async def delete_vfolders_forever(self, vfolder_ids: list[uuid.UUID]) -> list[VFolderData]:
        """
        Delete VFolders forever
        """

        async with self._db.connect() as db_conn:
            async with self._db.begin_session(db_conn) as db_session:
                vfolder_rows = []
                for vfolder_id in vfolder_ids:
                    vfolder_row = await self._get_vfolder_by_id(db_session, vfolder_id)
                    if vfolder_row:
                        vfolder_rows.append(vfolder_row)
                delete_stmt = (
                    sa.update(VFolderRow)
                    .where(VFolderRow.id.in_(vfolder_ids))
                    .values(status=VFolderOperationStatus.DELETE_ONGOING)
                )
                await db_session.execute(delete_stmt)

            # Delete relation rows
            await delete_vfolder_relation_rows(db_conn, self._db.begin_session, vfolder_ids)

            return [self._vfolder_row_to_data(row) for row in vfolder_rows]

    @repository_decorator()
    async def get_vfolder_permissions(self, vfolder_id: uuid.UUID) -> list[VFolderPermissionData]:
        """
        Get all permissions for a VFolder.
        """
        async with self._db.begin_session() as session:
            query = sa.select(VFolderPermissionRow).where(
                VFolderPermissionRow.vfolder == vfolder_id
            )
            result = await session.execute(query)
            permission_rows = result.scalars().all()

            return [
                VFolderPermissionData(
                    id=row.id,
                    vfolder=row.vfolder,
                    user=row.user,
                    permission=row.permission,
                )
                for row in permission_rows
            ]

    @repository_decorator()
    async def create_vfolder_permission(
        self,
        vfolder_id: uuid.UUID,
        user_id: uuid.UUID,
        permission: VFolderPermission,
    ) -> VFolderPermissionData:
        """
        Create a VFolder permission entry.
        """
        async with self._db.begin_session() as session:
            permission_id = uuid.uuid4()
            insert_values = {
                "id": permission_id,
                "vfolder": vfolder_id,
                "user": user_id,
                "permission": permission,
            }

            query = sa.insert(VFolderPermissionRow, insert_values)
            await session.execute(query)

            await self._role_manager.map_entity_to_scope(
                session,
                entity_id=ObjectId(
                    entity_type=EntityType.VFOLDER,
                    entity_id=str(vfolder_id),
                ),
                scope_id=ScopeId(ScopeType.USER, str(user_id)),
            )
            await self._role_manager.add_object_permission_to_user_role(
                session,
                user_id=user_id,
                entity_id=ObjectId(
                    entity_type=EntityType.VFOLDER,
                    entity_id=str(vfolder_id),
                ),
                operations=permission.to_rbac_operation(),
            )

            return VFolderPermissionData(
                id=permission_id,
                vfolder=vfolder_id,
                user=user_id,
                permission=permission,
            )

    @repository_decorator()
    async def delete_vfolder_permission(self, vfolder_id: uuid.UUID, user_id: uuid.UUID) -> None:
        """
        Delete a VFolder permission entry.
        """
        async with self._db.begin_session() as session:
            query = sa.delete(VFolderPermissionRow).where(
                (VFolderPermissionRow.vfolder == vfolder_id)
                & (VFolderPermissionRow.user == user_id)
            )
            await session.execute(query)
            await self._role_manager.unmap_entity_from_scope(
                session,
                entity_id=ObjectId(
                    entity_type=EntityType.VFOLDER,
                    entity_id=str(vfolder_id),
                ),
                scope_id=ScopeId(ScopeType.USER, str(user_id)),
            )
            await self._role_manager.delete_object_permission_of_user(
                session,
                user_id,
                vfolder_id,
            )

    @repository_decorator()
    async def get_vfolder_invitations_by_vfolder(
        self, vfolder_id: uuid.UUID
    ) -> list[VFolderInvitationData]:
        """
        Get all invitations for a VFolder.
        """
        async with self._db.begin_session() as session:
            query = sa.select(VFolderInvitationRow).where(
                VFolderInvitationRow.vfolder == vfolder_id
            )
            result = await session.execute(query)
            invitation_rows = result.scalars().all()

            return [
                VFolderInvitationData(
                    id=row.id,
                    vfolder=row.vfolder,
                    inviter=row.inviter,
                    invitee=row.invitee,
                    permission=row.permission,
                    created_at=row.created_at,
                    modified_at=row.modified_at,
                )
                for row in invitation_rows
            ]

    @repository_decorator()
    async def count_vfolders_by_user(self, user_id: uuid.UUID) -> int:
        """
        Count VFolders owned by a user (excluding hard deleted ones).
        """
        from ai.backend.manager.models.vfolder import HARD_DELETED_VFOLDER_STATUSES

        async with self._db.begin_session() as session:
            query = (
                sa.select(sa.func.count())
                .select_from(VFolderRow)
                .where(
                    (VFolderRow.user == user_id)
                    & (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES))
                )
            )
            result = await session.scalar(query)
            return result or 0

    @repository_decorator()
    async def count_vfolders_by_group(self, group_id: uuid.UUID) -> int:
        """
        Count VFolders owned by a group (excluding hard deleted ones).
        """
        from ai.backend.manager.models.vfolder import HARD_DELETED_VFOLDER_STATUSES

        async with self._db.begin_session() as session:
            query = (
                sa.select(sa.func.count())
                .select_from(VFolderRow)
                .where(
                    (VFolderRow.group == group_id)
                    & (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES))
                )
            )
            result = await session.scalar(query)
            return result or 0

    @repository_decorator()
    async def check_vfolder_name_exists(
        self,
        name: str,
        user_id: uuid.UUID,
        user_role: UserRole,
        domain_name: str,
        allowed_vfolder_types: list[str],
    ) -> bool:
        """
        Check if a VFolder with the given name already exists for the user.
        """
        from ai.backend.manager.models.vfolder import HARD_DELETED_VFOLDER_STATUSES

        async with self._db.begin_session() as session:
            # Use query_accessible_vfolders to check accessible folders
            extra_vf_conds = sa.and_(
                (VFolderRow.name == name),
                (VFolderRow.status.not_in(HARD_DELETED_VFOLDER_STATUSES)),
            )

            vfolder_dicts = await query_accessible_vfolders(
                session.bind,
                user_id,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=extra_vf_conds,
            )

            return len(vfolder_dicts) > 0

    @repository_decorator()
    async def get_user_info(self, user_id: uuid.UUID) -> Optional[tuple[UserRole, str]]:
        """
        Get user role and domain name for a user.
        Returns (role, domain_name) or None if user not found.
        """
        async with self._db.begin_session() as session:
            user_row = await session.scalar(sa.select(UserRow).where(UserRow.uuid == user_id))
            if not user_row:
                return None
            return user_row.role, user_row.domain_name

    @repository_decorator()
    async def get_user_email_by_id(self, user_id: uuid.UUID) -> Optional[str]:
        """
        Get user email by user ID.
        Returns email or None if user not found.
        """
        async with self._db.begin_session() as session:
            user_row = await session.scalar(sa.select(UserRow).where(UserRow.uuid == user_id))
            if not user_row:
                return None
            return user_row.email

    @repository_decorator()
    async def get_users_by_ids(self, user_ids: list[uuid.UUID]) -> list[tuple[uuid.UUID, str]]:
        """
        Get user info for multiple user IDs.
        Returns list of (user_id, email) tuples.
        """
        async with self._db.begin_session() as session:
            result = await session.execute(sa.select(UserRow).where(UserRow.uuid.in_(user_ids)))
            user_rows = result.scalars().all()
            return [(row.uuid, row.email) for row in user_rows]

    @repository_decorator()
    async def get_group_resource_info(
        self, group_id_or_name: Union[str, uuid.UUID], domain_name: str
    ) -> Optional[tuple[uuid.UUID, int, int, ProjectType]]:
        """
        Get group resource information by group ID or name.
        Returns (group_uuid, max_vfolder_count, max_quota_scope_size, group_type) or None.
        """

        async with self._db.begin_session() as session:
            if isinstance(group_id_or_name, str):
                query = (
                    sa.select(GroupRow)
                    .where(
                        (GroupRow.domain_name == domain_name) & (GroupRow.name == group_id_or_name)
                    )
                    .options(selectinload(GroupRow.resource_policy_row))
                )
            else:  # UUID
                query = (
                    sa.select(GroupRow)
                    .where(
                        (GroupRow.domain_name == domain_name) & (GroupRow.id == group_id_or_name)
                    )
                    .options(selectinload(GroupRow.resource_policy_row))
                )

            result = await session.execute(query)
            group_row = result.scalar()

            if not group_row:
                return None

            return (
                group_row.id,
                group_row.resource_policy_row.max_vfolder_count,
                group_row.resource_policy_row.max_quota_scope_size,
                group_row.type,
            )

    @repository_decorator()
    async def get_user_resource_info(
        self, user_id: uuid.UUID
    ) -> Optional[tuple[int, int, Optional[int]]]:
        """
        Get user resource information.
        Returns (max_vfolder_count, max_quota_scope_size, container_uid) or None.
        """
        async with self._db.begin_session() as session:
            query = (
                sa.select(UserRow)
                .where(UserRow.uuid == user_id)
                .options(selectinload(UserRow.resource_policy_row))
            )
            result = await session.execute(query)
            user_row = result.scalar()

            if not user_row:
                return None

            return (
                user_row.resource_policy_row.max_vfolder_count,
                user_row.resource_policy_row.max_quota_scope_size,
                user_row.container_uid,
            )

    async def _get_vfolder_by_id(
        self, session: SASession, vfolder_id: uuid.UUID
    ) -> Optional[VFolderRow]:
        """
        Private method to get a VFolder by ID using an existing session.
        """
        query = sa.select(VFolderRow).where(VFolderRow.id == vfolder_id)
        result = await session.execute(query)
        return result.scalar()

    async def _validate_vfolder_ownership(
        self, session: SASession, vfolder_id: uuid.UUID, user_id: uuid.UUID
    ) -> VFolderRow:
        """
        Private method to validate VFolder ownership.
        Raises VFolderNotFound if vfolder doesn't exist or user doesn't own it.
        """
        vfolder_row = await self._get_vfolder_by_id(session, vfolder_id)
        if not vfolder_row:
            raise VFolderNotFound()

        # Check ownership
        is_owner = False
        if vfolder_row.ownership_type == VFolderOwnershipType.USER:
            is_owner = vfolder_row.user == user_id
        elif vfolder_row.ownership_type == VFolderOwnershipType.GROUP:
            # TODO: check group membership
            pass

        if not is_owner:
            raise VFolderNotFound()

        return vfolder_row

    def _vfolder_row_to_data(self, row: VFolderRow) -> VFolderData:
        """
        Convert VFolderRow to VFolderData.
        """
        return VFolderData(
            id=row.id,
            name=row.name,
            host=row.host,
            domain_name=row.domain_name,
            quota_scope_id=row.quota_scope_id,
            usage_mode=row.usage_mode,
            permission=row.permission,
            max_files=row.max_files,
            max_size=row.max_size,
            num_files=row.num_files,
            cur_size=row.cur_size,
            created_at=row.created_at,
            last_used=row.last_used,
            creator=row.creator,
            unmanaged_path=row.unmanaged_path,
            ownership_type=row.ownership_type,
            user=row.user,
            group=row.group,
            cloneable=row.cloneable,
            status=row.status,
        )

    @repository_decorator()
    async def check_user_has_vfolder_permission(
        self, vfolder_id: uuid.UUID, user_ids: list[uuid.UUID]
    ) -> bool:
        """
        Check if any of the users already have permission for the vfolder.
        Returns True if any user already has permission.
        """
        from ai.backend.manager.models.vfolder import VFolderPermissionRow

        async with self._db.begin_session() as session:
            # Check direct permissions and ownership
            j = sa.join(
                VFolderPermissionRow, VFolderRow, VFolderRow.id == VFolderPermissionRow.vfolder
            )
            count_query = (
                sa.select(sa.func.count())
                .select_from(j)
                .where(
                    sa.and_(
                        sa.or_(
                            VFolderPermissionRow.user.in_(user_ids),
                            VFolderRow.user.in_(user_ids),
                        ),
                        VFolderPermissionRow.vfolder == vfolder_id,
                    )
                )
            )
            count = await session.scalar(count_query)
            return count > 0

    @repository_decorator()
    async def get_user_by_email(self, email: str) -> Optional[tuple[uuid.UUID, str]]:
        """
        Get user info by email.
        Returns (user_id, domain_name) or None if user not found.
        """
        async with self._db.begin_session() as session:
            user_row = await session.scalar(sa.select(UserRow).where(UserRow.email == email))
            if not user_row:
                return None
            return user_row.uuid, user_row.domain_name

    @repository_decorator()
    async def get_users_by_emails(self, emails: list[str]) -> list[tuple[uuid.UUID, str]]:
        """
        Get user info for multiple emails.
        Returns list of (user_id, email) tuples.
        """
        async with self._db.begin_session() as session:
            result = await session.execute(sa.select(UserRow).where(UserRow.email.in_(emails)))
            user_rows = result.scalars().all()
            return [(row.uuid, row.email) for row in user_rows]

    @repository_decorator()
    async def count_vfolder_with_name_for_user(self, user_id: uuid.UUID, vfolder_name: str) -> int:
        """
        Count VFolders with the given name accessible to the user.
        Used to check for duplicates when accepting invitations.
        """
        from ai.backend.manager.models.vfolder import VFolderStatusSet, vfolder_status_map

        async with self._db.begin_session() as session:
            j = sa.join(
                VFolderRow,
                VFolderPermissionRow,
                VFolderRow.id == VFolderPermissionRow.vfolder,
                isouter=True,
            )
            query = (
                sa.select(sa.func.count())
                .select_from(j)
                .where(
                    sa.and_(
                        sa.or_(
                            VFolderRow.user == user_id,
                            VFolderPermissionRow.user == user_id,
                        ),
                        VFolderRow.name == vfolder_name,
                        VFolderRow.status.not_in(vfolder_status_map[VFolderStatusSet.INACCESSIBLE]),
                    )
                )
            )
            result = await session.scalar(query)
            return result or 0

    @repository_decorator()
    async def check_pending_invitation_exists(
        self, vfolder_id: uuid.UUID, inviter_email: str, invitee_email: str
    ) -> bool:
        """
        Check if a pending invitation already exists.
        Returns True if a pending invitation exists.
        """
        async with self._db.begin_session() as session:
            query = (
                sa.select(sa.func.count())
                .select_from(VFolderInvitationRow)
                .where(
                    (VFolderInvitationRow.inviter == inviter_email)
                    & (VFolderInvitationRow.invitee == invitee_email)
                    & (VFolderInvitationRow.vfolder == vfolder_id)
                    & (VFolderInvitationRow.state == VFolderInvitationState.PENDING),
                )
            )
            result = await session.scalar(query)
            return result > 0

    @repository_decorator()
    async def create_vfolder_invitation(
        self,
        vfolder_id: uuid.UUID,
        inviter_email: str,
        invitee_email: str,
        permission: VFolderPermission,
    ) -> Optional[str]:
        """
        Create a VFolder invitation.
        Returns the invitee email on success, None on failure.
        """
        from sqlalchemy import exc as sa_exc

        async with self._db.begin_session() as session:
            query = sa.insert(
                VFolderInvitationRow,
                {
                    "permission": permission,
                    "vfolder": vfolder_id,
                    "inviter": inviter_email,
                    "invitee": invitee_email,
                    "state": VFolderInvitationState.PENDING,
                },
            )
            try:
                await session.execute(query)
                return invitee_email
            except sa_exc.DataError:
                return None

    @repository_decorator()
    async def get_invitation_by_id(
        self, invitation_id: uuid.UUID
    ) -> Optional[VFolderInvitationData]:
        """
        Get invitation by ID.
        Returns VFolderInvitationData or None if not found.
        """
        async with self._db.begin_session() as session:
            query = sa.select(VFolderInvitationRow).where(
                (VFolderInvitationRow.id == invitation_id)
                & (VFolderInvitationRow.state == VFolderInvitationState.PENDING),
            )
            invitation_row = await session.scalar(query)
            if not invitation_row:
                return None

            return VFolderInvitationData(
                id=invitation_row.id,
                vfolder=invitation_row.vfolder,
                inviter=invitation_row.inviter,
                invitee=invitation_row.invitee,
                permission=invitation_row.permission,
                created_at=invitation_row.created_at,
                modified_at=invitation_row.modified_at,
            )

    @repository_decorator()
    async def update_invitation_state(
        self, invitation_id: uuid.UUID, new_state: VFolderInvitationState
    ) -> None:
        """
        Update invitation state.
        """
        async with self._db.begin_session() as session:
            query = (
                sa.update(VFolderInvitationRow)
                .where(VFolderInvitationRow.id == invitation_id)
                .values(state=new_state)
            )
            await session.execute(query)

    @repository_decorator()
    async def update_invitation_permission(
        self, invitation_id: uuid.UUID, inviter_email: str, permission: VFolderPermission
    ) -> None:
        """
        Update invitation permission (only by inviter).
        """
        async with self._db.begin_session() as session:
            query = (
                sa.update(VFolderInvitationRow)
                .values(permission=permission)
                .where(
                    sa.and_(
                        VFolderInvitationRow.id == invitation_id,
                        VFolderInvitationRow.inviter == inviter_email,
                        VFolderInvitationRow.state == VFolderInvitationState.PENDING,
                    )
                )
            )
            await session.execute(query)

    @repository_decorator()
    async def update_invited_vfolder_mount_permission(
        self, vfolder_id: uuid.UUID, user_id: uuid.UUID, permission: VFolderPermission
    ) -> None:
        """
        Update the permission of an invited user for a specific vfolder.
        """
        async with self._db.begin_session() as session:
            query = (
                sa.update(VFolderPermissionRow)
                .where(
                    sa.and_(
                        VFolderPermissionRow.vfolder == vfolder_id,
                        VFolderPermissionRow.user == user_id,
                    )
                )
                .values(permission=permission)
            )
            await session.execute(query)

    @repository_decorator()
    async def get_pending_invitations_for_user(
        self, user_email: str
    ) -> list[tuple[VFolderInvitationData, VFolderData]]:
        """
        Get all pending invitations for a user with VFolder info.
        Returns list of (invitation_data, vfolder_data) tuples.
        """
        from sqlalchemy.orm import contains_eager

        async with self._db.begin_session() as session:
            j = sa.join(
                VFolderInvitationRow, VFolderRow, VFolderInvitationRow.vfolder == VFolderRow.id
            )
            query = (
                sa.select(VFolderInvitationRow)
                .select_from(j)
                .where(
                    sa.and_(
                        VFolderInvitationRow.invitee == user_email,
                        VFolderInvitationRow.state == VFolderInvitationState.PENDING,
                    )
                )
                .options(
                    contains_eager(VFolderInvitationRow.vfolder_row),
                )
            )
            invitation_rows = await session.scalars(query)
            invitation_rows = invitation_rows.all()

            results = []
            for inv_row in invitation_rows:
                invitation_data = VFolderInvitationData(
                    id=inv_row.id,
                    vfolder=inv_row.vfolder,
                    inviter=inv_row.inviter,
                    invitee=inv_row.invitee,
                    permission=inv_row.permission,
                    created_at=inv_row.created_at,
                    modified_at=inv_row.modified_at,
                )
                vfolder_data = self._vfolder_row_to_data(inv_row.vfolder_row)
                results.append((invitation_data, vfolder_data))

            return results

    @repository_decorator()
    async def ensure_host_permission_allowed(
        self,
        folder_host: str,
        *,
        permission: VFolderHostPermission,
        allowed_vfolder_types: Sequence[str],
        user_uuid: uuid.UUID,
        resource_policy: Mapping[str, Any],
        domain_name: str,
        group_id: Optional[uuid.UUID] = None,
    ) -> None:
        """
        Ensure that the user has the required permission on the specified vfolder host.
        """
        async with self._db.begin_session() as session:
            await ensure_host_permission_allowed(
                session.bind,
                folder_host,
                permission=permission,
                allowed_vfolder_types=allowed_vfolder_types,
                user_uuid=user_uuid,
                resource_policy=resource_policy,
                domain_name=domain_name,
                group_id=group_id,
            )

    def _vfolder_dict_to_data(self, vfolder_dict: dict[str, Any]) -> VFolderData:
        """
        Convert vfolder dictionary from query_accessible_vfolders to VFolderData.
        """
        return VFolderData(
            id=vfolder_dict["id"],
            name=vfolder_dict["name"],
            host=vfolder_dict["host"],
            domain_name=vfolder_dict["domain_name"],
            quota_scope_id=vfolder_dict["quota_scope_id"],
            usage_mode=vfolder_dict["usage_mode"],
            permission=vfolder_dict.get("permission"),
            max_files=vfolder_dict["max_files"],
            max_size=vfolder_dict["max_size"],
            num_files=vfolder_dict.get("num_files", 0),
            cur_size=vfolder_dict["cur_size"],
            created_at=vfolder_dict["created_at"],
            last_used=vfolder_dict["last_used"],
            creator=vfolder_dict["creator"],
            unmanaged_path=vfolder_dict["unmanaged_path"],
            ownership_type=vfolder_dict["ownership_type"],
            user=uuid.UUID(vfolder_dict["user"]) if vfolder_dict["user"] else None,
            group=uuid.UUID(vfolder_dict["group"]) if vfolder_dict["group"] else None,
            cloneable=vfolder_dict["cloneable"],
            status=vfolder_dict["status"],
        )

    @repository_decorator()
    async def initiate_vfolder_clone(
        self,
        vfolder_info: VFolderCloneInfo,
        storage_manager: StorageSessionManager,
        background_task_manager: BackgroundTaskManager,
    ) -> tuple[uuid.UUID, uuid.UUID]:
        """
        Initiate VFolder cloning process.
        Returns (task_id, target_folder_id).
        """
        source_vf_cond = vfolders.c.id == vfolder_info.source_vfolder_id.folder_id

        async def _update_source_status() -> None:
            async with self._db.begin_session() as db_session:
                query = (
                    sa.update(vfolders)
                    .values(status=VFolderOperationStatus.CLONING)
                    .where(source_vf_cond)
                )
                await db_session.execute(query)

        await execute_with_retry(_update_source_status)

        target_proxy, target_volume = storage_manager.get_proxy_and_volume(vfolder_info.target_host)
        source_proxy, source_volume = storage_manager.get_proxy_and_volume(
            vfolder_info.source_host, is_unmanaged(vfolder_info.unmanaged_path)
        )

        if source_proxy != target_proxy:
            raise VFolderInvalidParameter(
                f"Proxy names of source and target vfolders must be equal. "
                f"Source proxy: {source_proxy}, Target proxy: {target_proxy}."
            )

        # Generate the ID of the destination vfolder.
        # TODO: If we refactor to use ORM, the folder ID will be created from the database by inserting
        #       the actual object (with RETURNING clause).  In that case, we need to temporarily
        #       mark the object to be "unusable-yet" until the storage proxy creates the destination
        #       vfolder.  After done, we need to make another transaction to clear the unusable state.
        target_folder_id = VFolderID(vfolder_info.source_vfolder_id.quota_scope_id, uuid.uuid4())

        # Clone the vfolder contents
        manager_client = storage_manager.get_manager_facing_client(source_proxy)
        clone_response = await manager_client.clone_folder(
            source_volume,
            str(vfolder_info.source_vfolder_id),
            target_volume,
            str(target_folder_id),
        )
        task_id = clone_response.bgtask_id

        async def _insert_vfolder() -> None:
            async with self._db.begin_session() as db_session:
                insert_values = {
                    "id": target_folder_id.folder_id,
                    "name": vfolder_info.target_vfolder_name,
                    "domain_name": vfolder_info.domain_name,
                    "usage_mode": vfolder_info.usage_mode,
                    "permission": vfolder_info.permission,
                    "last_used": None,
                    "host": vfolder_info.target_host,
                    # TODO: add quota_scope_id
                    "creator": vfolder_info.email,
                    "ownership_type": VFolderOwnershipType("user"),
                    "user": vfolder_info.user_id,
                    "group": None,
                    "unmanaged_path": None,
                    "cloneable": vfolder_info.cloneable,
                    "quota_scope_id": vfolder_info.source_vfolder_id.quota_scope_id,
                }
                query = sa.insert(vfolders).values(**insert_values)
                await db_session.execute(query)

        # Insert the new vfolder record
        await execute_with_retry(_insert_vfolder)

        return task_id, target_folder_id.folder_id

    @repository_decorator()
    async def get_logs_vfolder(
        self,
        user_id: uuid.UUID,
        user_role: UserRole,
        domain_name: str,
    ) -> Optional[VFolderData]:
        """
        Get the accessible .logs vfolder for a user.
        Returns VFolderData if found, None otherwise.
        """
        async with self._db.begin_readonly() as conn:
            vfolder_dicts = await query_accessible_vfolders(
                conn,
                user_id,
                user_role=user_role,
                domain_name=domain_name,
                allowed_vfolder_types=["user"],
                extra_vf_conds=(vfolders.c.name == ".logs"),
            )

            if not vfolder_dicts:
                return None

            # Return the first (and should be only) matching .logs vfolder
            vfolder_dict = vfolder_dicts[0]
            return self._vfolder_dict_to_data(dict(vfolder_dict))



"""Deployment service for managing model deployments."""

import logging

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.services.deployment.actions.create_deployment import (
    CreateDeploymentAction,
    CreateDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.destroy_deployment import (
    DestroyDeploymentAction,
    DestroyDeploymentActionResult,
)
from ai.backend.manager.sokovan.deployment import DeploymentController
from ai.backend.manager.sokovan.deployment.types import DeploymentLifecycleType

log = BraceStyleAdapter(logging.getLogger(__name__))


class DeploymentService:
    """Service for managing deployments."""

    _deployment_controller: DeploymentController

    def __init__(self, deployment_controller: DeploymentController) -> None:
        """Initialize deployment service with controller."""
        self._deployment_controller = deployment_controller

    async def create(self, action: CreateDeploymentAction) -> CreateDeploymentActionResult:
        """Create a new deployment.

        Args:
            action: Create deployment action containing the creator specification

        Returns:
            CreateDeploymentActionResult: Result containing the created deployment info
        """
        log.info("Creating deployment with name: {}", action.creator.name)
        deployment_info = await self._deployment_controller.create_deployment(action.creator)
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_PENDING
        )
        return CreateDeploymentActionResult(data=deployment_info)

    async def destroy(self, action: DestroyDeploymentAction) -> DestroyDeploymentActionResult:
        """Destroy an existing deployment.

        Args:
            action: Destroy deployment action containing the endpoint ID

        Returns:
            DestroyDeploymentActionResult: Result indicating success or failure
        """
        log.info("Destroying deployment with ID: {}", action.endpoint_id)
        success = await self._deployment_controller.destroy_deployment(action.endpoint_id)
        await self._deployment_controller.mark_lifecycle_needed(DeploymentLifecycleType.DESTROYING)
        return DestroyDeploymentActionResult(success=success)



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.types import RepositoryArgs
from ai.backend.manager.repositories.vfolder.repository import VfolderRepository


@dataclass
class VfolderRepositories:
    repository: VfolderRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = VfolderRepository(args.db)

        return cls(
            repository=repository,
        )



"""set null to empty unmanaged_path field

Revision ID: 01f0ed604819
Revises: 17ea44123131
Create Date: 2025-07-02 16:31:50.692652

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "01f0ed604819"
down_revision = "17ea44123131"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("""
    UPDATE vfolders
    SET unmanaged_path = NULL
    WHERE unmanaged_path = '';
    """)


def downgrade() -> None:
    # Downgrade is not implemented as the operation is irreversible.
    pass



"""Add `availability` column to `artifacts` table

Revision ID: 08c1867e266e
Revises: 982a7c19a06c
Create Date: 2025-09-11 03:21:04.096533

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "08c1867e266e"
down_revision = "982a7c19a06c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "artifacts", sa.Column("availability", sa.String(), nullable=False, server_default="ALIVE")
    )
    op.create_index(op.f("ix_artifacts_availability"), "artifacts", ["availability"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_artifacts_availability"), table_name="artifacts")
    op.drop_column("artifacts", "availability")
    # ### end Alembic commands ###



"""Handler for provisioning routes."""

import logging
from collections.abc import Sequence
from typing import Optional

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.model_serving.types import RouteStatus
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.deployment.types import RouteData
from ai.backend.manager.sokovan.deployment.route.executor import RouteExecutor
from ai.backend.manager.sokovan.deployment.route.types import RouteExecutionResult

from .base import RouteHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class ProvisioningRouteHandler(RouteHandler):
    """Handler for provisioning routes (PROVISIONING -> UNHEALTHY)."""

    def __init__(
        self,
        route_executor: RouteExecutor,
        event_producer: EventProducer,
    ):
        self._route_executor = route_executor
        self._event_producer = event_producer

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "provision-routes"

    @property
    def lock_id(self) -> Optional[LockID]:
        """Lock for provisioning routes."""
        return LockID.LOCKID_DEPLOYMENT_PROVISIONING_ROUTES

    @classmethod
    def target_statuses(cls) -> list[RouteStatus]:
        """Get the target route statuses for this handler."""
        return [RouteStatus.PROVISIONING]

    @classmethod
    def next_status(cls) -> Optional[RouteStatus]:
        """Get the next route status after this handler's operation."""
        return RouteStatus.UNHEALTHY

    @classmethod
    def failure_status(cls) -> Optional[RouteStatus]:
        """Get the failure route status if applicable."""
        return RouteStatus.FAILED_TO_START

    async def execute(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Execute provisioning for routes."""
        log.debug("Provisioning {} routes", len(routes))

        # Execute route provisioning logic via executor
        result = await self._route_executor.provision_routes(routes)
        return result

    async def post_process(self, result: RouteExecutionResult) -> None:
        """Handle post-processing after provisioning routes."""
        log.info(
            "Provisioned {} routes successfully, {} failed",
            len(result.successes),
            len(result.errors),
        )



"""Handler for terminating sessions."""

import logging
from typing import Optional

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
from ai.backend.manager.sokovan.scheduler.results import ScheduleResult
from ai.backend.manager.sokovan.scheduler.scheduler import Scheduler
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

from .base import SchedulerHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class TerminateSessionsHandler(SchedulerHandler):
    """Handler for terminating sessions."""

    def __init__(
        self,
        scheduler: Scheduler,
        scheduling_controller: SchedulingController,
        event_producer: EventProducer,
        repository: SchedulerRepository,
    ):
        self._scheduler = scheduler
        self._scheduling_controller = scheduling_controller
        self._event_producer = event_producer
        self._repository = repository

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "terminate-sessions"

    @property
    def lock_id(self) -> Optional[LockID]:
        """Lock for operations targeting TERMINATING sessions."""
        return LockID.LOCKID_SOKOVAN_TARGET_TERMINATING

    async def execute(self) -> ScheduleResult:
        """Terminate sessions marked for termination."""
        return await self._scheduler.terminate_sessions()

    async def post_process(self, result: ScheduleResult) -> None:
        """
        No post-processing needed.

        Actual status updates and events are handled by:
        - Agent event callbacks (for successful terminations)
        - sweep_lost_agent_kernels() (for lost agents or failed RPC calls)
        """
        # No action needed - terminate_sessions only sends RPC calls
        pass



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.services.container_registry.actions.base import ContainerRegistryAction


@dataclass
class LoadAllContainerRegistriesAction(ContainerRegistryAction):
    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "load_all"


@dataclass
class LoadAllContainerRegistriesActionResult(BaseActionResult):
    registries: list[ContainerRegistryData]

    @override
    def entity_id(self) -> Optional[str]:
        return None



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.resource_preset.actions.base import ResourcePresetAction


@dataclass
class ListResourcePresetsAction(ResourcePresetAction):
    access_key: str
    scaling_group: Optional[str]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "list_multi"


@dataclass
class ListResourcePresetsResult(BaseActionResult):
    # TODO: Add preset type
    presets: list[Any]

    @override
    def entity_id(self) -> Optional[str]:
        # TODO: Should return preset row ids after changing to batching.
        return None



"""add_unique_constraints_to_association_tables

Revision ID: a5319bfc7d7c
Revises: caf54fcc17ab
Create Date: 2024-01-03 21:43:31.208183

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID, mapper_registry

# revision identifiers, used by Alembic.
revision = "a5319bfc7d7c"
down_revision = "caf54fcc17ab"
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()

    association_groups_users = sa.Table(
        "association_groups_users",
        mapper_registry.metadata,
        sa.Column(
            "id",
            GUID,
            primary_key=True,
            nullable=False,
        ),
        sa.Column(
            "user_id",
            GUID,
            sa.ForeignKey("users.uuid", onupdate="CASCADE", ondelete="CASCADE"),
            nullable=False,
        ),
        sa.Column(
            "group_id",
            GUID,
            sa.ForeignKey("groups.id", onupdate="CASCADE", ondelete="CASCADE"),
            nullable=False,
        ),
        sa.UniqueConstraint("user_id", "group_id", name="uq_association_user_id_group_id"),
        extend_existing=True,
    )

    sgroups_for_domains = sa.Table(
        "sgroups_for_domains",
        mapper_registry.metadata,
        sa.Column(
            "id",
            GUID,
            primary_key=True,
            nullable=False,
        ),
        sa.Column(
            "scaling_group",
            sa.ForeignKey("scaling_groups.name", onupdate="CASCADE", ondelete="CASCADE"),
            index=True,
            nullable=False,
        ),
        sa.Column(
            "domain",
            sa.ForeignKey("domains.name", onupdate="CASCADE", ondelete="CASCADE"),
            index=True,
            nullable=False,
        ),
        extend_existing=True,
    )

    sgroups_for_groups = sa.Table(
        "sgroups_for_groups",
        mapper_registry.metadata,
        sa.Column(
            "id",
            GUID,
            primary_key=True,
            nullable=False,
        ),
        sa.Column(
            "scaling_group",
            sa.ForeignKey("scaling_groups.name", onupdate="CASCADE", ondelete="CASCADE"),
            index=True,
            nullable=False,
        ),
        sa.Column(
            "group",
            sa.ForeignKey("groups.id", onupdate="CASCADE", ondelete="CASCADE"),
            index=True,
            nullable=False,
        ),
        extend_existing=True,
    )

    sgroups_for_keypairs = sa.Table(
        "sgroups_for_keypairs",
        mapper_registry.metadata,
        sa.Column(
            "id",
            GUID,
            primary_key=True,
            nullable=False,
        ),
        sa.Column(
            "scaling_group",
            sa.ForeignKey("scaling_groups.name", onupdate="CASCADE", ondelete="CASCADE"),
            index=True,
            nullable=False,
        ),
        sa.Column(
            "access_key",
            sa.ForeignKey("keypairs.access_key", onupdate="CASCADE", ondelete="CASCADE"),
            index=True,
            nullable=False,
        ),
        extend_existing=True,
    )

    def ensure_unique(table, field_1: str, field_2: str) -> None:
        # Leave only one duplicate record and delete all of it
        t1 = table.alias("t1")
        t2 = table.alias("t2")
        subq = (
            sa.select([t1.c.id])
            .where(t1.c[field_1] == t2.c[field_1])
            .where(t1.c[field_2] == t2.c[field_2])
            .where(t1.c.id > t2.c.id)
        )
        delete_stmt = sa.delete(table).where(table.c.id.in_(subq))
        conn.execute(delete_stmt)

    ensure_unique(association_groups_users, "user_id", "group_id")
    ensure_unique(sgroups_for_domains, "scaling_group", "domain")
    ensure_unique(sgroups_for_groups, "scaling_group", "group")
    ensure_unique(sgroups_for_keypairs, "scaling_group", "access_key")

    op.create_unique_constraint(
        "uq_association_user_id_group_id", "association_groups_users", ["user_id", "group_id"]
    )
    op.create_unique_constraint(
        "uq_sgroup_domain", "sgroups_for_domains", ["scaling_group", "domain"]
    )
    op.create_unique_constraint(
        "uq_sgroup_ugroup", "sgroups_for_groups", ["scaling_group", "group"]
    )
    op.create_unique_constraint(
        "uq_sgroup_akey", "sgroups_for_keypairs", ["scaling_group", "access_key"]
    )


def downgrade():
    op.drop_constraint(
        "uq_association_user_id_group_id", "association_groups_users", type_="unique"
    )
    op.drop_constraint("uq_sgroup_domain", "sgroups_for_domains", type_="unique")
    op.drop_constraint("uq_sgroup_ugroup", "sgroups_for_groups", type_="unique")
    op.drop_constraint("uq_sgroup_akey", "sgroups_for_keypairs", type_="unique")



"""Scheduling controller for managing session lifecycle and scheduling operations."""

import logging
from dataclasses import dataclass

from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.contexts.user import current_user
from ai.backend.common.defs import RESERVED_VFOLDER_PATTERNS, RESERVED_VFOLDERS
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.session.broadcast import SchedulingBroadcastEvent
from ai.backend.common.events.types import AbstractBroadcastEvent
from ai.backend.common.exception import InvalidAPIParameters
from ai.backend.common.plugin.hook import ALL_COMPLETED, PASSED, HookPluginContext
from ai.backend.common.types import ResourceSlot, SessionId
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.errors.common import RejectedByHook
from ai.backend.manager.metrics.scheduler import (
    SchedulerOperationMetricObserver,
    SchedulerPhaseMetricObserver,
)
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.plugin.network import NetworkPluginContext
from ai.backend.manager.repositories.scheduler import (
    MarkTerminatingResult,
    SchedulerRepository,
)
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    AllowedScalingGroup,
    SessionCreationSpec,
)
from ai.backend.manager.scheduler.types import ScheduleType
from ai.backend.manager.sokovan.scheduling_controller.types import SessionValidationSpec

from .calculators.resource_calculator import ResourceCalculator
from .preparers import (
    ClusterConfigurationRule,
    InternalDataRule,
    MountPreparationRule,
    SessionPreparer,
)
from .resolvers.scaling_group_resolver import ScalingGroupResolver
from .validators import (
    ClusterValidationRule,
    ContainerLimitRule,
    MountNameValidationRule,
    ScalingGroupAccessRule,
    ServicePortRule,
    SessionTypeRule,
    SessionValidator,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class SchedulingControllerArgs:
    """Arguments for initializing SchedulingController."""

    repository: SchedulerRepository
    config_provider: ManagerConfigProvider
    storage_manager: StorageSessionManager
    event_producer: EventProducer
    valkey_schedule: ValkeyScheduleClient
    network_plugin_ctx: NetworkPluginContext
    hook_plugin_ctx: HookPluginContext


class SchedulingController:
    """Controller for session lifecycle and scheduling operations management."""

    _repository: SchedulerRepository
    _config_provider: ManagerConfigProvider
    _storage_manager: StorageSessionManager
    _event_producer: EventProducer
    _valkey_schedule: ValkeyScheduleClient
    _network_plugin_ctx: NetworkPluginContext

    # Services
    _scaling_group_resolver: ScalingGroupResolver
    _validator: SessionValidator
    _preparer: SessionPreparer
    _resource_calculator: ResourceCalculator
    _metric_observer: SchedulerPhaseMetricObserver
    _operation_metrics: SchedulerOperationMetricObserver
    _hook_plugin_ctx: HookPluginContext

    def __init__(self, args: SchedulingControllerArgs) -> None:
        """Initialize the scheduling controller with required services."""
        self._repository = args.repository
        self._config_provider = args.config_provider
        self._storage_manager = args.storage_manager
        self._event_producer = args.event_producer
        self._valkey_schedule = args.valkey_schedule
        self._network_plugin_ctx = args.network_plugin_ctx
        self._hook_plugin_ctx = args.hook_plugin_ctx

        # Initialize metric observers (singletons)
        self._metric_observer = SchedulerPhaseMetricObserver.instance()
        self._operation_metrics = SchedulerOperationMetricObserver.instance()

        # Initialize services
        self._scaling_group_resolver = ScalingGroupResolver()

        # Initialize validator with rules
        validator_rules = [
            ContainerLimitRule(),
            ScalingGroupAccessRule(),
            SessionTypeRule(),
            ServicePortRule(),
            ClusterValidationRule(),
            MountNameValidationRule(),
        ]
        self._validator = SessionValidator(validator_rules)

        # Initialize preparer with rules
        preparer_rules = [
            ClusterConfigurationRule(),
            MountPreparationRule(),
            InternalDataRule(),
        ]
        self._preparer = SessionPreparer(preparer_rules)

        # Initialize resource calculator (still needed for resource calculations)
        self._resource_calculator = ResourceCalculator(args.config_provider)

    async def _resolve_scaling_group(
        self,
        session_spec: SessionCreationSpec,
    ) -> AllowedScalingGroup:
        """
        Resolve the scaling group for the session.

        If scaling group is specified in spec, use it.
        Otherwise, fetch allowed groups and auto-select.

        Args:
            session_spec: Session creation specification

        Returns:
            str: The resolved scaling group name
        """
        # Fetch allowed groups to determine the scaling group
        allowed_groups = await self._repository.query_allowed_scaling_groups(
            session_spec.user_scope.domain_name,
            str(session_spec.user_scope.group_id),
            session_spec.access_key,
        )
        if session_spec.scaling_group:
            for sg in allowed_groups:
                if sg.name == session_spec.scaling_group:
                    return sg
            raise InvalidAPIParameters(
                f"Scaling group '{session_spec.scaling_group}' is not accessible"
            )

        # Resolve the scaling group
        return self._scaling_group_resolver.resolve(
            session_spec,
            allowed_groups,
        )

    async def enqueue_session(
        self,
        session_spec: SessionCreationSpec,
    ) -> SessionId:
        """
        Enqueue a new session for scheduling.

        Steps:
        1. Resolve scaling group
        2. Fetch all required data from repository
        3. Validate the specification
        4. Calculate resources and prepare session data
        5. Enqueue in repository

        Args:
            session_spec: Session creation specification

        Returns:
            SessionId: The ID of the created session
        """
        # Phase 1: Resolve scaling group
        with self._metric_observer.measure_phase(
            "scheduling_controller", "", "resolve_scaling_group"
        ):
            validated_scaling_group = await self._resolve_scaling_group(session_spec)

        # Phase 2: Fetch all required data
        with self._metric_observer.measure_phase(
            "scheduling_controller", validated_scaling_group.name, "fetch_data"
        ):
            allowed_vfolder_types = list(
                await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
            )
            creation_context = await self._repository.fetch_session_creation_data(
                session_spec,
                validated_scaling_group.name,
                self._storage_manager,
                allowed_vfolder_types,
            )

        # Phase 3: Validate
        with self._metric_observer.measure_phase(
            "scheduling_controller", validated_scaling_group.name, "validation"
        ):
            self._validator.validate(
                session_spec,
                creation_context,
            )

        # Phase 4: Calculate resources and prepare session data
        with self._metric_observer.measure_phase(
            "scheduling_controller", validated_scaling_group.name, "preparation"
        ):
            # Pre-calculate resources
            calculated_resources = await self._resource_calculator.calculate(
                validated_scaling_group,
                session_spec,
                creation_context,
            )

            # Prepare session data with calculated resources
            session_data = await self._preparer.prepare(
                session_spec,
                validated_scaling_group,
                creation_context,
                calculated_resources,
            )

        hook_result = await self._hook_plugin_ctx.dispatch(
            "PRE_ENQUEUE_SESSION",
            (session_data.id, session_data.name, session_data.access_key),
            return_when=ALL_COMPLETED,
        )
        if hook_result.status != PASSED:
            raise RejectedByHook.from_hook_result(hook_result)

        # Phase 5: Enqueue in repository
        with self._metric_observer.measure_phase(
            "scheduling_controller", validated_scaling_group.name, "enqueue"
        ):
            session_id = await self._repository.enqueue_session(session_data)

        log.info(
            "Session {} ({}) enqueued successfully",
            session_data.name,
            session_id,
        )
        try:
            await self.mark_scheduling_needed(ScheduleType.SCHEDULE)
        except Exception as e:
            log.warning(
                "Failed to request scheduling for session {}: {}",
                session_id,
                e,
            )
        await self._hook_plugin_ctx.notify(
            "POST_ENQUEUE_SESSION",
            (session_id, session_data.name, session_data.access_key),
        )
        return session_id

    async def mark_scheduling_needed(self, schedule_type: ScheduleType) -> None:
        """
        Request a scheduling operation for the next cycle.

        This is the public interface for requesting scheduling operations.
        The actual scheduling will be handled internally by the coordinator.

        Args:
            schedule_type: Type of scheduling to request
        """
        await self._valkey_schedule.mark_schedule_needed(schedule_type.value)
        log.debug("Requested scheduling for type: {}", schedule_type.value)

    async def mark_sessions_for_termination(
        self,
        session_ids: list[SessionId],
        reason: str = "USER_REQUESTED",
    ) -> MarkTerminatingResult:
        """
        Mark multiple sessions and their kernels for termination by updating their status to TERMINATING.

        This method handles the lifecycle management of sessions by marking them
        for termination, which will be processed by the scheduler's terminate_sessions method.
        It also automatically requests TERMINATE scheduling if sessions were processed.

        Args:
            session_ids: List of session IDs to terminate
            reason: Reason for termination

        Returns:
            MarkTerminatingResult with categorized session statuses
        """
        result = await self._repository.mark_sessions_terminating(session_ids, reason)

        if result.has_processed():
            log.info(
                "Marked {} sessions for termination (cancelled: {}, terminating: {})",
                result.processed_count(),
                len(result.cancelled_sessions),
                len(result.terminating_sessions),
            )

            cancelled_events: list[AbstractBroadcastEvent] = [
                SchedulingBroadcastEvent(
                    session_id=session_id,
                    creation_id="",
                    status_transition=str(SessionStatus.CANCELLED),
                    reason=reason,
                )
                for session_id in result.cancelled_sessions
            ]
            if cancelled_events:
                await self._event_producer.broadcast_events_batch(cancelled_events)
            # Record metric for termination attempts
            self._operation_metrics.observe_success(
                operation="mark_sessions_terminating",
                count=result.processed_count(),
            )
            # Request termination scheduling for the next cycle
            await self.mark_scheduling_needed(ScheduleType.TERMINATE)

        return result

    async def validate_session_spec(self, spec: SessionValidationSpec) -> None:
        # TODO: Refactor to use ValidationRule
        alias_folders = spec.mount_spec.mount_map.values()
        if len(alias_folders) != len(set(alias_folders)):
            raise InvalidAPIParameters("Duplicate alias folder name exists.")
        original_folders = spec.mount_spec.mount_map.keys()
        alias_name: str
        for alias_name in alias_folders:
            if alias_name.startswith("/home/work/"):
                alias_name = alias_name.replace("/home/work/", "")
            if alias_name == "":
                raise InvalidAPIParameters("Alias name cannot be empty.")
            if not _verify_vfolder_name(alias_name):
                raise InvalidAPIParameters(str(alias_name) + " is reserved for internal path.")
            if alias_name in original_folders:
                raise InvalidAPIParameters(
                    "Alias name cannot be set to an existing folder name: " + str(alias_name)
                )
        # Validate resource slots
        available_resource_slots = (
            await self._config_provider.legacy_etcd_config_loader.get_resource_slots()
        )
        try:
            ResourceSlot.from_user_input(
                spec.resource_spec.resource_slots, available_resource_slots
            )
        except ValueError as e:
            raise InvalidAPIParameters(f"Invalid resource allocation: {e}")
        # Validate Image
        user = current_user()
        if user is None:
            raise InvalidAPIParameters("User context is required for image validation.")
        await self._repository.check_available_image(
            spec.image_identifier, user.domain_name, user.user_id
        )


def _verify_vfolder_name(folder: str) -> bool:
    if folder in RESERVED_VFOLDERS:
        return False
    for pattern in RESERVED_VFOLDER_PATTERNS:
        if pattern.match(folder):
            return False
    return True



import enum


class BasePermission(enum.StrEnum):
    pass


class VFolderPermission(BasePermission):
    # Only owners can do
    CLONE = enum.auto()
    ASSIGN_PERMISSION_TO_OTHERS = enum.auto()  # Invite, share

    # `create_vfolder` action should be in {Domain, Project, or User} permissions, not here
    READ_ATTRIBUTE = enum.auto()
    UPDATE_ATTRIBUTE = enum.auto()
    DELETE_VFOLDER = enum.auto()

    READ_CONTENT = enum.auto()
    WRITE_CONTENT = enum.auto()
    DELETE_CONTENT = enum.auto()

    MOUNT_RO = enum.auto()
    MOUNT_RW = enum.auto()
    MOUNT_WD = enum.auto()  # Legacy


class StorageHostPermission(BasePermission):
    CREATE_FOLDER = enum.auto()

    CLONE = VFolderPermission.CLONE
    ASSIGN_PERMISSION_TO_OTHERS = VFolderPermission.ASSIGN_PERMISSION_TO_OTHERS

    READ_ATTRIBUTE = VFolderPermission.READ_ATTRIBUTE
    UPDATE_ATTRIBUTE = VFolderPermission.UPDATE_ATTRIBUTE
    DELETE_VFOLDER = VFolderPermission.DELETE_VFOLDER

    READ_CONTENT = VFolderPermission.READ_CONTENT
    WRITE_CONTENT = VFolderPermission.WRITE_CONTENT
    DELETE_CONTENT = VFolderPermission.DELETE_CONTENT

    MOUNT_RO = VFolderPermission.MOUNT_RO
    MOUNT_RW = VFolderPermission.MOUNT_RW
    MOUNT_WD = VFolderPermission.MOUNT_WD


class ComputeSessionPermission(BasePermission):
    # `create_session` action should be in {Domain, Project, or User} permissions, not here
    READ_ATTRIBUTE = enum.auto()
    UPDATE_ATTRIBUTE = enum.auto()
    DELETE_SESSION = enum.auto()

    START_APP = enum.auto()
    EXECUTE = enum.auto()
    CONVERT_TO_IMAGE = enum.auto()


class ScalingGroupPermission(BasePermission):
    # super-admin only
    ASSOCIATE_WITH_SCOPES = enum.auto()
    ASSIGN_AGENTS = enum.auto()
    UPDATE_ATTRIBUTE = enum.auto()

    # admin only
    READ_ATTRIBUTE = enum.auto()

    # Permission set of bindings and sub-scopes
    AGENT_PERMISSIONS = enum.auto()
    COMPUTE_SESSION_PERMISSIONS = enum.auto()
    INFERENCE_SERVICE_PERMISSIONS = enum.auto()

    STORAGE_HOST_PERMISSIONS = enum.auto()


class AgentPermission(BasePermission):
    READ_ATTRIBUTE = enum.auto()
    UPDATE_ATTRIBUTE = enum.auto()

    CREATE_COMPUTE_SESSION = enum.auto()
    CREATE_SERVICE = enum.auto()


class DomainPermission(BasePermission):
    # These permissions limit actions taken directly to domains
    READ_ATTRIBUTE = enum.auto()
    READ_SENSITIVE_ATTRIBUTE = enum.auto()
    UPDATE_ATTRIBUTE = enum.auto()

    CREATE_USER = enum.auto()
    CREATE_PROJECT = enum.auto()


class ProjectPermission(BasePermission):
    # These permissions limit actions taken directly to projects(groups)
    READ_ATTRIBUTE = enum.auto()
    READ_SENSITIVE_ATTRIBUTE = enum.auto()
    UPDATE_ATTRIBUTE = enum.auto()
    DELETE_PROJECT = enum.auto()

    ASSOCIATE_WITH_USER = enum.auto()


class ImagePermission(BasePermission):
    READ_ATTRIBUTE = enum.auto()
    UPDATE_ATTRIBUTE = enum.auto()
    CREATE_CONTAINER = enum.auto()

    FORGET_IMAGE = enum.auto()



import logging
from datetime import datetime, timedelta
from typing import Optional, Sequence
from uuid import UUID

from dateutil.relativedelta import relativedelta

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import (
    InvalidAPIParameters,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.models.resource_usage import (
    ProjectResourceUsage,
    parse_resource_usage_groups,
    parse_total_resource_group,
)
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.repositories.group.admin_repository import AdminGroupRepository
from ai.backend.manager.repositories.group.repositories import GroupRepositories
from ai.backend.manager.repositories.group.repository import GroupRepository
from ai.backend.manager.services.group.actions.create_group import (
    CreateGroupAction,
    CreateGroupActionResult,
)
from ai.backend.manager.services.group.actions.delete_group import (
    DeleteGroupAction,
    DeleteGroupActionResult,
)
from ai.backend.manager.services.group.actions.modify_group import (
    ModifyGroupAction,
    ModifyGroupActionResult,
)
from ai.backend.manager.services.group.actions.purge_group import (
    PurgeGroupAction,
    PurgeGroupActionResult,
)
from ai.backend.manager.services.group.actions.usage_per_month import (
    UsagePerMonthAction,
    UsagePerMonthActionResult,
)
from ai.backend.manager.services.group.actions.usage_per_period import (
    UsagePerPeriodAction,
    UsagePerPeriodActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GroupService:
    _config_provider: ManagerConfigProvider
    _valkey_stat_client: ValkeyStatClient
    _storage_manager: StorageSessionManager
    _group_repository: GroupRepository
    _admin_group_repository: AdminGroupRepository

    def __init__(
        self,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
        valkey_stat_client: ValkeyStatClient,
        group_repositories: GroupRepositories,
    ) -> None:
        self._storage_manager = storage_manager
        self._config_provider = config_provider
        self._valkey_stat_client = valkey_stat_client
        self._group_repository = group_repositories.repository
        self._admin_group_repository = group_repositories.admin_repository

    async def create_group(self, action: CreateGroupAction) -> CreateGroupActionResult:
        group_data = await self._group_repository.create(action.input)
        return CreateGroupActionResult(data=group_data)

    async def modify_group(self, action: ModifyGroupAction) -> ModifyGroupActionResult:
        from ai.backend.manager.models.user import UserRole

        # Convert user_uuids from list[str] to list[UUID] if provided
        user_uuids_converted = None
        user_uuids_list = action.user_uuids.optional_value()
        if user_uuids_list:
            user_uuids_converted = [UUID(user_uuid) for user_uuid in user_uuids_list]

        group_data = await self._group_repository.modify_validated(
            action.group_id,
            action.modifier,
            UserRole.USER,  # Default role since group operations don't require role-based logic
            action.user_update_mode.optional_value(),
            user_uuids_converted,
        )
        # If no group data is returned, it means only user updates were performed
        return ModifyGroupActionResult(data=group_data)

    async def delete_group(self, action: DeleteGroupAction) -> DeleteGroupActionResult:
        await self._group_repository.mark_inactive(action.group_id)
        return DeleteGroupActionResult(group_id=action.group_id)

    async def purge_group(self, action: PurgeGroupAction) -> PurgeGroupActionResult:
        await self._admin_group_repository.purge_group_force(action.group_id)
        return PurgeGroupActionResult(group_id=action.group_id)

    async def _get_project_stats_for_period(
        self,
        start_date: datetime,
        end_date: datetime,
        project_ids: Optional[Sequence[UUID]] = None,
    ) -> dict[UUID, ProjectResourceUsage]:
        kernels = await self._group_repository.fetch_project_resource_usage(
            start_date, end_date, project_ids=project_ids
        )
        local_tz = self._config_provider.config.system.timezone
        usage_groups = await parse_resource_usage_groups(
            kernels, self._valkey_stat_client, local_tz
        )
        total_groups, _ = parse_total_resource_group(usage_groups)
        return total_groups

    # group (or all the groups)
    async def usage_per_month(self, action: UsagePerMonthAction) -> UsagePerMonthActionResult:
        month = action.month
        local_tz = self._config_provider.config.system.timezone

        try:
            start_date = datetime.strptime(month, "%Y%m").replace(tzinfo=local_tz)
            end_date = start_date + relativedelta(months=+1)
        except ValueError:
            raise InvalidAPIParameters(extra_msg="Invalid date values")
        result = await self._group_repository.get_container_stats_for_period(
            start_date, end_date, action.group_ids
        )
        log.debug("container list are retrieved for month {0}", month)
        return UsagePerMonthActionResult(result=result)

    # group (or all the groups)
    async def usage_per_period(self, action: UsagePerPeriodAction) -> UsagePerPeriodActionResult:
        local_tz = self._config_provider.config.system.timezone
        project_id = action.project_id

        try:
            start_date = datetime.strptime(action.start_date, "%Y%m%d").replace(tzinfo=local_tz)
            end_date = datetime.strptime(action.end_date, "%Y%m%d").replace(tzinfo=local_tz)
            end_date = end_date + timedelta(days=1)  # include sessions in end_date
            if end_date - start_date > timedelta(days=100):
                raise InvalidAPIParameters("Cannot query more than 100 days")
        except ValueError:
            raise InvalidAPIParameters(extra_msg="Invalid date values")
        if end_date <= start_date:
            raise InvalidAPIParameters(extra_msg="end_date must be later than start_date.")
        log.info(
            "USAGE_PER_MONTH (p:{}, start_date:{}, end_date:{})", project_id, start_date, end_date
        )
        project_ids = [project_id] if project_id is not None else None
        usage_map = await self._get_project_stats_for_period(
            start_date, end_date, project_ids=project_ids
        )
        result = [p_usage.to_json(child=True) for p_usage in usage_map.values()]
        log.debug("container list are retrieved from {0} to {1}", start_date, end_date)
        return UsagePerPeriodActionResult(result=result)



from __future__ import annotations

import uuid
from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Self,
    Sequence,
)

import graphene
import graphql
import sqlalchemy as sa
from dateutil.parser import parse as dtparse
from graphene.types.datetime import DateTime as GQLDateTime
from graphql import Undefined
from sqlalchemy.engine.row import Row

from ai.backend.common.exception import (
    InvalidAPIParameters,
)
from ai.backend.common.types import ResourceSlot
from ai.backend.manager.data.group.types import GroupCreator, GroupData, GroupModifier
from ai.backend.manager.models.rbac import ProjectScope
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.group.actions.create_group import CreateGroupAction
from ai.backend.manager.services.group.actions.delete_group import (
    DeleteGroupAction,
)
from ai.backend.manager.services.group.actions.modify_group import ModifyGroupAction
from ai.backend.manager.services.group.actions.purge_group import (
    PurgeGroupAction,
)
from ai.backend.manager.types import OptionalState, TriState

from ..base import (
    BigInt,
    FilterExprArg,
    OrderExprArg,
    PaginatedConnectionField,
    batch_multiresult,
    batch_result,
    generate_sql_info_for_gql_connection,
    privileged_mutation,
)
from ..gql_relay import (
    AsyncNode,
    Connection,
    ConnectionResolverResult,
)
from ..group import (
    AssocGroupUserRow,
    GroupRow,
    ProjectType,
    association_groups_users,
    get_permission_ctx,
    groups,
)
from ..minilang.ordering import OrderSpecItem, QueryOrderParser
from ..minilang.queryfilter import FieldSpecItem, QueryFilterParser
from ..rbac.context import ClientContext
from ..rbac.permission_defs import ProjectPermission
from .user import UserConnection, UserNode

if TYPE_CHECKING:
    from ..gql import GraphQueryContext
    from ..rbac import ContainerRegistryScope, ScopeType
    from .scaling_group import ScalingGroup


__all__ = (
    "GroupNode",
    "GroupConnection",
    "Group",
    "GroupInput",
    "ModifyGroupInput",
    "CreateGroup",
    "ModifyGroup",
    "DeleteGroup",
    "PurgeGroup",
    "GroupPermissionField",
)


class GroupNode(graphene.ObjectType):
    class Meta:
        interfaces = (AsyncNode,)

    row_id = graphene.UUID(description="Added in 24.03.7. The undecoded id value stored in DB.")
    name = graphene.String()
    description = graphene.String()
    is_active = graphene.Boolean()
    created_at = GQLDateTime()
    modified_at = GQLDateTime()
    domain_name = graphene.String()
    total_resource_slots = graphene.JSONString()
    allowed_vfolder_hosts = graphene.JSONString()
    integration_id = graphene.String()
    resource_policy = graphene.String()
    type = graphene.String(description=f"Added in 24.03.7. One of {[t.name for t in ProjectType]}.")
    container_registry = graphene.JSONString(description="Added in 24.03.7.")
    scaling_groups = graphene.List(
        lambda: graphene.String,
    )

    registry_quota = BigInt(description="Added in 25.3.0.")

    user_nodes = PaginatedConnectionField(
        UserConnection,
    )

    queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
        "id": ("id", None),
        "name": ("name", None),
        "is_active": ("is_active", None),
        "created_at": ("created_at", dtparse),
        "modified_at": ("modified_at", dtparse),
        "domain_name": ("domain_name", None),
        "resource_policy": ("resource_policy", None),
    }

    queryorder_colmap: Mapping[str, OrderSpecItem] = {
        "id": ("id", None),
        "name": ("name", None),
        "is_active": ("is_active", None),
        "created_at": ("created_at", None),
        "modified_at": ("modified_at", None),
        "domain_name": ("domain_name", None),
        "resource_policy": ("resource_policy", None),
    }

    @classmethod
    def from_row(
        cls,
        graph_ctx: GraphQueryContext,
        row: GroupRow,
    ) -> Self:
        return cls(
            id=row.id,
            row_id=row.id,
            name=row.name,
            description=row.description,
            is_active=row.is_active,
            created_at=row.created_at,
            modified_at=row.modified_at,
            domain_name=row.domain_name,
            total_resource_slots=row.total_resource_slots.to_json() or {},
            allowed_vfolder_hosts=row.allowed_vfolder_hosts.to_json() or {},
            integration_id=row.integration_id,
            resource_policy=row.resource_policy,
            type=row.type.name,
            container_registry=row.container_registry,
        )

    async def resolve_scaling_groups(self, info: graphene.ResolveInfo) -> Sequence[ScalingGroup]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(
            graph_ctx,
            "ScalingGroup.by_group",
        )
        sgroups = await loader.load(self.id)
        return [sg.name for sg in sgroups]

    async def resolve_user_nodes(
        self,
        info: graphene.ResolveInfo,
        filter: str | None = None,
        order: str | None = None,
        offset: int | None = None,
        after: str | None = None,
        first: int | None = None,
        before: str | None = None,
        last: int | None = None,
    ) -> ConnectionResolverResult[Self]:
        from ..user import UserRow

        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter, QueryFilterParser(UserNode._queryfilter_fieldspec))
            if filter is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order, QueryOrderParser(UserNode._queryorder_colmap))
            if order is not None
            else None
        )
        (
            query,
            _,
            conditions,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            UserRow,
            UserRow.uuid,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )
        j = sa.join(UserRow, AssocGroupUserRow)
        user_query = query.select_from(j).where(AssocGroupUserRow.group_id == self.id)
        cnt_query = (
            sa.select(sa.func.count()).select_from(j).where(AssocGroupUserRow.group_id == self.id)
        )
        for cond in conditions:
            cnt_query = cnt_query.where(cond)
        async with graph_ctx.db.begin_readonly_session() as db_session:
            user_rows = (await db_session.scalars(user_query)).all()
            result = [type(self).from_row(graph_ctx, row) for row in user_rows]
            total_cnt = await db_session.scalar(cnt_query)
            return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)

    async def resolve_registry_quota(self, info: graphene.ResolveInfo) -> int:
        graph_ctx: GraphQueryContext = info.context
        scope_id = ProjectScope(project_id=self.id, domain_name=None)

        return await graph_ctx.services_ctx.per_project_container_registries_quota.read_quota(
            scope_id,
        )

    @classmethod
    async def get_node(cls, info: graphene.ResolveInfo, id) -> Self:
        graph_ctx: GraphQueryContext = info.context
        _, group_id = AsyncNode.resolve_global_id(info, id)
        query = sa.select(GroupRow).where(GroupRow.id == group_id)
        async with graph_ctx.db.begin_readonly_session() as db_session:
            group_row = (await db_session.scalars(query)).first()
            return cls.from_row(graph_ctx, group_row)

    @classmethod
    async def get_connection(
        cls,
        info: graphene.ResolveInfo,
        scope: ScopeType,
        container_registry_scope: Optional[ContainerRegistryScope] = None,
        permission: ProjectPermission = ProjectPermission.READ_ATTRIBUTE,
        filter_expr: Optional[str] = None,
        order_expr: Optional[str] = None,
        offset: Optional[int] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        before: Optional[str] = None,
        last: Optional[int] = None,
    ) -> ConnectionResolverResult[Self]:
        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter_expr, QueryFilterParser(cls.queryfilter_fieldspec))
            if filter_expr is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order_expr, QueryOrderParser(cls.queryorder_colmap))
            if order_expr is not None
            else None
        )
        (
            query,
            cnt_query,
            _,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            GroupRow,
            GroupRow.id,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )
        async with graph_ctx.db.connect() as db_conn:
            user = graph_ctx.user
            client_ctx = ClientContext(
                graph_ctx.db, user["domain_name"], user["uuid"], user["role"]
            )
            permission_ctx = await get_permission_ctx(
                db_conn, client_ctx, permission, scope, container_registry_scope
            )
            cond = permission_ctx.query_condition
            if cond is None:
                return ConnectionResolverResult([], cursor, pagination_order, page_size, 0)
            query = query.where(cond)
            cnt_query = cnt_query.where(cond)

            async with graph_ctx.db.begin_readonly_session(db_conn) as db_session:
                group_rows = (await db_session.scalars(query)).all()
                total_cnt = await db_session.scalar(cnt_query)
                result = [cls.from_row(graph_ctx, row) for row in group_rows]

        return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)


class GroupConnection(Connection):
    class Meta:
        node = GroupNode
        description = "Added in 24.03.0"


class GroupPermissionField(graphene.Scalar):
    class Meta:
        description = f"Added in 25.3.0. One of {[val.value for val in ProjectPermission]}."

    @staticmethod
    def serialize(val: ProjectPermission) -> str:
        return val.value

    @staticmethod
    def parse_literal(node: Any, _variables=None):
        if isinstance(node, graphql.language.ast.StringValueNode):
            return ProjectPermission(node.value)

    @staticmethod
    def parse_value(value: str) -> ProjectPermission:
        return ProjectPermission(value)


class Group(graphene.ObjectType):
    id = graphene.UUID()
    name = graphene.String()
    description = graphene.String()
    is_active = graphene.Boolean()
    created_at = GQLDateTime()
    modified_at = GQLDateTime()
    domain_name = graphene.String()
    total_resource_slots = graphene.JSONString()
    allowed_vfolder_hosts = graphene.JSONString()
    integration_id = graphene.String()
    resource_policy = graphene.String()
    type = graphene.String(description="Added in 24.03.0.")
    container_registry = graphene.JSONString(description="Added in 24.03.0.")

    scaling_groups = graphene.List(lambda: graphene.String)

    @classmethod
    def from_row(cls, graph_ctx: GraphQueryContext, row: Row) -> Optional[Group]:
        if row is None:
            return None
        return cls(
            id=row["id"],
            name=row["name"],
            description=row["description"],
            is_active=row["is_active"],
            created_at=row["created_at"],
            modified_at=row["modified_at"],
            domain_name=row["domain_name"],
            total_resource_slots=(
                row["total_resource_slots"].to_json()
                if row["total_resource_slots"] is not None
                else {}
            ),
            allowed_vfolder_hosts=row["allowed_vfolder_hosts"].to_json(),
            integration_id=row["integration_id"],
            resource_policy=row["resource_policy"],
            type=row["type"].name,
            container_registry=row["container_registry"],
        )

    @classmethod
    def from_dto(cls, dto: Optional[GroupData]) -> Optional[Self]:
        if dto is None:
            return None
        return cls(
            id=dto.id,
            name=dto.name,
            description=dto.description,
            is_active=dto.is_active,
            created_at=dto.created_at,
            modified_at=dto.modified_at,
            domain_name=dto.domain_name,
            total_resource_slots=dto.total_resource_slots.to_json()
            if dto.total_resource_slots
            else {},
            allowed_vfolder_hosts=dto.allowed_vfolder_hosts.to_json(),
            integration_id=dto.integration_id,
            resource_policy=dto.resource_policy,
            type=dto.type.name,
            container_registry=dto.container_registry,
        )

    async def resolve_scaling_groups(self, info: graphene.ResolveInfo) -> Sequence[ScalingGroup]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(
            graph_ctx,
            "ScalingGroup.by_group",
        )
        sgroups = await loader.load(self.id)
        return [sg.name for sg in sgroups]

    @classmethod
    async def load_all(
        cls,
        graph_ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        is_active: Optional[bool] = None,
        type: list[ProjectType] = [ProjectType.GENERAL],
    ) -> Sequence[Group]:
        query = sa.select([groups]).select_from(groups).where(groups.c.type.in_(type))
        if domain_name is not None:
            query = query.where(groups.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(groups.c.is_active == is_active)
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, row)) is not None
            ]

    @classmethod
    async def batch_load_by_id(
        cls,
        graph_ctx: GraphQueryContext,
        group_ids: Sequence[uuid.UUID],
        *,
        domain_name: Optional[str] = None,
        is_active: Optional[bool] = None,
    ) -> Sequence[Group | None]:
        query = sa.select([groups]).select_from(groups).where(groups.c.id.in_(group_ids))
        if domain_name is not None:
            query = query.where(groups.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(groups.c.is_active == is_active)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_result(
                graph_ctx,
                conn,
                query,
                cls,
                group_ids,
                lambda row: row["id"],
            )

    @classmethod
    async def batch_load_by_name(
        cls,
        graph_ctx: GraphQueryContext,
        group_names: Sequence[str],
        *,
        domain_name: Optional[str] = None,
        is_active: Optional[bool] = None,
    ) -> Sequence[Sequence[Group | None]]:
        query = sa.select([groups]).select_from(groups).where(groups.c.name.in_(group_names))
        if domain_name is not None:
            query = query.where(groups.c.domain_name == domain_name)
        if is_active is not None:
            query = query.where(groups.c.is_active == is_active)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_multiresult(
                graph_ctx,
                conn,
                query,
                cls,
                group_names,
                lambda row: row["name"],
            )

    @classmethod
    async def batch_load_by_user(
        cls,
        graph_ctx: GraphQueryContext,
        user_ids: Sequence[uuid.UUID],
        *,
        type: list[ProjectType] | None = None,
        is_active: Optional[bool] = None,
    ) -> Sequence[Sequence[Group | None]]:
        if type is None:
            _type = [ProjectType.GENERAL]
        else:
            _type = type
        j = sa.join(
            groups,
            association_groups_users,
            groups.c.id == association_groups_users.c.group_id,
        )
        query = (
            sa.select([groups, association_groups_users.c.user_id])
            .select_from(j)
            .where(association_groups_users.c.user_id.in_(user_ids) & (groups.c.type.in_(_type)))
        )
        if is_active is not None:
            query = query.where(groups.c.is_active == is_active)
        async with graph_ctx.db.begin_readonly() as conn:
            return await batch_multiresult(
                graph_ctx,
                conn,
                query,
                cls,
                user_ids,
                lambda row: row["user_id"],
            )

    @classmethod
    async def get_groups_for_user(
        cls,
        graph_ctx: GraphQueryContext,
        user_id: uuid.UUID,
    ) -> Sequence[Group]:
        j = sa.join(
            groups,
            association_groups_users,
            groups.c.id == association_groups_users.c.group_id,
        )
        query = (
            sa.select([groups]).select_from(j).where(association_groups_users.c.user_id == user_id)
        )
        async with graph_ctx.db.begin_readonly() as conn:
            return [
                obj
                async for row in (await conn.stream(query))
                if (obj := cls.from_row(graph_ctx, row)) is not None
            ]


class GroupInput(graphene.InputObjectType):
    type = graphene.String(
        required=False,
        default_value="GENERAL",
        description=(
            f"Added in 24.03.0. Available values: {', '.join([p.name for p in ProjectType])}"
        ),
    )
    description = graphene.String(required=False, default_value="")
    is_active = graphene.Boolean(required=False, default_value=True)
    domain_name = graphene.String(required=True)
    total_resource_slots = graphene.JSONString(required=False, default_value={})
    allowed_vfolder_hosts = graphene.JSONString(required=False, default_value={})
    integration_id = graphene.String(required=False, default_value="")
    resource_policy = graphene.String(required=False, default_value="default")
    container_registry = graphene.JSONString(
        required=False, default_value={}, description="Added in 24.03.0"
    )

    def to_action(self, name: str) -> CreateGroupAction:
        def value_or_none(value):
            return value if value is not Undefined else None

        type_val = None if self.type is Undefined else ProjectType[self.type]
        description_val = value_or_none(self.description)
        is_active_val = value_or_none(self.is_active)
        total_resource_slots_val = (
            None
            if self.total_resource_slots is Undefined
            else ResourceSlot.from_user_input(self.total_resource_slots, None)
        )
        allowed_vfolder_hosts_val = value_or_none(self.allowed_vfolder_hosts)
        integration_id_val = value_or_none(self.integration_id)
        resource_policy_val = value_or_none(self.resource_policy)
        container_registry_val = value_or_none(self.container_registry)

        return CreateGroupAction(
            input=GroupCreator(
                name=name,
                domain_name=self.domain_name,
                type=type_val,
                description=description_val,
                is_active=is_active_val,
                total_resource_slots=total_resource_slots_val,
                allowed_vfolder_hosts=allowed_vfolder_hosts_val,
                integration_id=integration_id_val,
                resource_policy=resource_policy_val,
                container_registry=container_registry_val,
            ),
        )


class ModifyGroupInput(graphene.InputObjectType):
    name = graphene.String(required=False)
    description = graphene.String(required=False)
    is_active = graphene.Boolean(required=False)
    domain_name = graphene.String(required=False)
    total_resource_slots = graphene.JSONString(required=False)
    user_update_mode = graphene.String(required=False)
    user_uuids = graphene.List(lambda: graphene.String, required=False)
    allowed_vfolder_hosts = graphene.JSONString(required=False)
    integration_id = graphene.String(required=False)
    resource_policy = graphene.String(required=False)
    container_registry = graphene.JSONString(
        required=False, default_value={}, description="Added in 24.03.0"
    )

    def to_action(self, group_id: uuid.UUID) -> ModifyGroupAction:
        return ModifyGroupAction(
            group_id=group_id,
            modifier=GroupModifier(
                name=OptionalState[str].from_graphql(
                    self.name,
                ),
                domain_name=OptionalState[str].from_graphql(
                    self.domain_name,
                ),
                description=TriState[str].from_graphql(
                    self.description,
                ),
                is_active=OptionalState[bool].from_graphql(
                    self.is_active,
                ),
                total_resource_slots=OptionalState[ResourceSlot].from_graphql(
                    self.total_resource_slots
                    if (self.total_resource_slots is Undefined or self.total_resource_slots is None)
                    else ResourceSlot.from_user_input(self.total_resource_slots, None),
                ),
                allowed_vfolder_hosts=OptionalState[dict[str, str]].from_graphql(
                    self.allowed_vfolder_hosts,
                ),
                integration_id=OptionalState[str].from_graphql(
                    self.integration_id,
                ),
                resource_policy=OptionalState[str].from_graphql(
                    self.resource_policy,
                ),
                container_registry=TriState[dict[str, str]].from_graphql(
                    self.container_registry,
                ),
            ),
            user_update_mode=OptionalState[str].from_graphql(
                self.user_update_mode,
            ),
            user_uuids=OptionalState[list[str]].from_graphql(
                self.user_uuids,
            ),
        )


class CreateGroup(graphene.Mutation):
    allowed_roles = (UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        name = graphene.String(required=True)
        props = GroupInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    group = graphene.Field(lambda: Group, required=False)

    @classmethod
    @privileged_mutation(
        UserRole.ADMIN,
        lambda name, props, **kwargs: (props.domain_name, None),
    )
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        name: str,
        props: GroupInput,
    ) -> CreateGroup:
        graph_ctx: GraphQueryContext = info.context
        if name.strip() == "" or len(name) > 64:
            raise InvalidAPIParameters(
                "Group name cannot be empty or whitespace and must not exceed 64 characters."
            )

        action = props.to_action(name)
        res = await graph_ctx.processors.group.create_group.wait_for_complete(action)
        return cls(
            ok=True,
            msg="success",
            group=Group.from_dto(res.data),
        )


class ModifyGroup(graphene.Mutation):
    allowed_roles = (UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        gid = graphene.UUID(required=True)
        props = ModifyGroupInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    group = graphene.Field(lambda: Group, required=False)

    @classmethod
    @privileged_mutation(
        UserRole.ADMIN,
        lambda gid, **kwargs: (None, gid),
    )
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        gid: uuid.UUID,
        props: ModifyGroupInput,
    ) -> ModifyGroup:
        graph_ctx: GraphQueryContext = info.context

        action = props.to_action(gid)
        res = await graph_ctx.processors.group.modify_group.wait_for_complete(action)
        return cls(
            ok=True,
            msg="success",
            group=Group.from_dto(res.data) if res.data else None,
        )


class DeleteGroup(graphene.Mutation):
    """
    Instead of deleting the group, just mark it as inactive.
    """

    allowed_roles = (UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        gid = graphene.UUID(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    @privileged_mutation(
        UserRole.ADMIN,
        lambda gid, **kwargs: (None, gid),
    )
    async def mutate(cls, root, info: graphene.ResolveInfo, gid: uuid.UUID) -> DeleteGroup:
        ctx: GraphQueryContext = info.context
        await ctx.processors.group.delete_group.wait_for_complete(DeleteGroupAction(gid))
        return cls(ok=True, msg="success")


class PurgeGroup(graphene.Mutation):
    """
    Completely deletes a group from DB.

    Group's vfolders and their data will also be lost
    as well as the kernels run from the group.
    There is no migration of the ownership for group folders.
    """

    allowed_roles = (UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        gid = graphene.UUID(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    @privileged_mutation(
        UserRole.ADMIN,
        lambda gid, **kwargs: (None, gid),
    )
    async def mutate(cls, root, info: graphene.ResolveInfo, gid: uuid.UUID) -> PurgeGroup:
        graph_ctx: GraphQueryContext = info.context

        await graph_ctx.processors.group.purge_group.wait_for_complete(PurgeGroupAction(gid))
        return cls(ok=True, msg="success")
